/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __AP_STREAM_PROTO_H__
#define __AP_STREAM_PROTO_H__

// All elements are transmitted in network byte order!

// How far off the sequence window can be
#define STREAM_SEQ_WINDOW    10

// Did we lose our way?
#define STREAM_SENTINEL      0xDECAFBAD

// Frames we speak, in no particular order other than what I thought of them in
#define FRAMETYPE_ANNOUNCE   0
#define FRAMETYPE_HELO       1
#define FRAMETYPE_CONFIGURE  2
#define FRAMETYPE_BANUSER    3
#define FRAMETYPE_MESSAGE    4
#define FRAMETYPE_ACTIVATE   5
#define FRAMETYPE_CHBSSID    6
#define FRAMETYPE_KISNETWORK 7
#define FRAMETYPE_KISALERT   8
#define FRAMETYPE_SMITE      9
#define FRAMETYPE_TERMINATE  254

// Length of frames
#define FRAMELEN_HEADER      (sizeof(struct stream_frame) - sizeof(void *))
#define FRAMELEN_HELO        (sizeof(struct ap_helo_packet))
#define FRAMELEN_ANNOUNCE    (sizeof(struct ap_announce_packet))
#define FRAMELEN_MESSAGE_P   (sizeof(struct ap_message_packet) - sizeof(uint8_t))
#define FRAMELEN_CONFIGURE   (sizeof(struct mn_configure_packet))
#define FRAMELEN_BANUSER     (sizeof(struct mn_banuser_packet))
#define FRAMELEN_ACTIVATE    (sizeof(struct mn_activate_packet))
#define FRAMELEN_CHBSSID     (sizeof(struct mn_bssid_packet))
#define FRAMELEN_KISALERT    (sizeof(struct ap_kisalert_packet))
#define FRAMELEN_KISNETWORK  (sizeof(struct ap_kisnetwork_packet))
#define FRAMELEN_SMITE       (sizeof(struct mn_smite_packet))
#define FRAMELEN_TERMINATE   0

// Interface control flags for announce and configure
#define AP_IFFLAG_NONE         0
// Interface is available
#define AP_IFFLAG_AVAILABLE    1
// Interface is (to be) an AP
#define AP_IFFLAG_ACCESSPOINT  2
// Interface is (to be) a sniffer
#define AP_IFFLAG_SNIFFER      4
// Interface is (to be) a rawtx engine
#define AP_IFFLAG_RAWTX        8
// Interface is configured
#define AP_IFFLAG_CONFIGURED   32
// Interface is dirty and needs to be reconfigured
#define AP_IFFLAG_DIRTY        64
// We're in an error state
#define AP_IFFLAG_ERROR        128
// We're busy working on something (like attacking)
#define AP_IFFLAG_BUSY         256
// Common conglomerates
#define AP_ACTIVEAP            (AP_IFFLAG_AVAILABLE | AP_IFFLAG_ACCESSPOINT | \
                                AP_IFFLAG_CONFIGURED)

// Message flags
#define AP_MSGFLAG_NONE    0
#define AP_MSGFLAG_DEBUG   1
#define AP_MSGFLAG_INFO    2
#define AP_MSGFLAG_ERROR   4
#define AP_MSGFLAG_FATAL   8

// Common frame header
typedef struct stream_frame {
    uint32_t frame_sentinel __attribute__ ((packed));
    uint8_t frame_type __attribute__ ((packed));
    uint32_t frame_len __attribute__ ((packed));
    void *payload;
};

// Common payload header so we can alias it in and handle auto-stuff like
// incrementing the sequence id
typedef struct ap_common_payload {
    uint16_t station_id __attribute__ ((packed));
    uint64_t sequence_id __attribute__ ((packed));
};

// Announce the entry of an AP into the network
typedef struct ap_announce_packet {
    // Station and sequence ID - not used in announce frames but we need them
    // for now anyhow
    uint16_t station_id __attribute__ ((packed));
    uint64_t sequence_id __attribute__ ((packed));
    // Name of the AP - not as important
    uint8_t name[32] __attribute__ ((packed));

    // Current BSSID - could be innate, could be one we pushed before
    uint8_t bssid0[6] __attribute__ ((packed));
    uint8_t bssid1[6] __attribute__ ((packed));
    uint8_t bssid2[6] __attribute__ ((packed));
    uint8_t bssid3[6] __attribute__ ((packed));

    // Interface names
    uint8_t ifname0[32] __attribute__ ((packed));
    uint8_t ifname1[32] __attribute__ ((packed));
    uint8_t ifname2[32] __attribute__ ((packed));
    uint8_t ifname3[32] __attribute__ ((packed));

    // Interface info
    uint32_t cf_bitmap0 __attribute__ ((packed));
    uint32_t cf_bitmap1 __attribute__ ((packed));
    uint32_t cf_bitmap2 __attribute__ ((packed));
    uint32_t cf_bitmap3 __attribute__ ((packed));

    // Zone number for the AP
    uint8_t zone __attribute__ ((packed));
};

typedef struct ap_helo_packet {
    // Common station and sequence IDs
    uint16_t station_id __attribute__ ((packed));
    uint64_t sequence_id __attribute__ ((packed));

    // Station-common stuff
    uint32_t station_timestamp_sec __attribute__ ((packed));
    uint32_t station_timestamp_usec __attribute__ ((packed));

    // Load average split into major.minor
    uint8_t load_avg_major __attribute__ ((packed));
    uint8_t load_avg_minor __attribute__ ((packed));

    // Per-interface stuff - this could be condensed into repeating 
    // substructs but for our purposes we'll cheat and do it this way
    
    // Associated clients
    uint32_t clientcount0 __attribute__ ((packed));
    uint32_t clientcount1 __attribute__ ((packed));
    uint32_t clientcount2 __attribute__ ((packed));
    uint32_t clientcount3 __attribute__ ((packed));

    // Frame counts per interface
    uint64_t framecount0 __attribute__ ((packed));
    uint64_t framecount1 __attribute__ ((packed));
    uint64_t framecount2 __attribute__ ((packed));
    uint64_t framecount3 __attribute__ ((packed));

    // Error counts per interface
    uint64_t errorcount0 __attribute__ ((packed));
    uint64_t errorcount1 __attribute__ ((packed));
    uint64_t errorcount2 __attribute__ ((packed));
    uint64_t errorcount3 __attribute__ ((packed));

    // Packet and byte counts per interface
    uint64_t txpacketcount0 __attribute__ ((packed));
    uint64_t txpacketcount1 __attribute__ ((packed));
    uint64_t txpacketcount2 __attribute__ ((packed));
    uint64_t txpacketcount3 __attribute__ ((packed));

    uint64_t rxpacketcount0 __attribute__ ((packed));
    uint64_t rxpacketcount1 __attribute__ ((packed));
    uint64_t rxpacketcount2 __attribute__ ((packed));
    uint64_t rxpacketcount3 __attribute__ ((packed));

    uint64_t txbytecount0 __attribute__ ((packed));
    uint64_t txbytecount1 __attribute__ ((packed));
    uint64_t txbytecount2 __attribute__ ((packed));
    uint64_t txbytecount3 __attribute__ ((packed));

    uint64_t rxbytecount0 __attribute__ ((packed));
    uint64_t rxbytecount1 __attribute__ ((packed));
    uint64_t rxbytecount2 __attribute__ ((packed));
    uint64_t rxbytecount3 __attribute__ ((packed));

    // Interface state/status bitmaps
    uint32_t cf_bitmap0 __attribute__ ((packed));
    uint32_t cf_bitmap1 __attribute__ ((packed));
    uint32_t cf_bitmap2 __attribute__ ((packed));
    uint32_t cf_bitmap3 __attribute__ ((packed));
};

// Send an arbitrary string message 
typedef struct ap_message_packet {
    // common
    uint16_t station_id __attribute__ ((packed));
    uint64_t sequence_id __attribute__ ((packed));
    // Flags - same as messagebus
    uint32_t msgflags;
    // Length of string
    uint32_t msglen;
    // Character array
    uint8_t message;
};

// Relay an alert to the central server
typedef struct ap_kisalert_packet {
    // common
    uint16_t station_id __attribute__ ((packed));
    uint64_t sequence_id __attribute__ ((packed));
    // Alert
    uint8_t type;
    uint8_t bssid_mac[6];
    uint8_t source_mac[6];
    uint8_t dest_mac[6];
    uint8_t other_mac[6];
    uint16_t channel;
    uint32_t timestamp;
};

// Relay a network to the central server
typedef struct ap_kisnetwork_packet {
    // common
    uint16_t station_id __attribute__ ((packed));
    uint64_t sequence_id __attribute__ ((packed));
    // network
    uint8_t bssid_mac[6];
    uint8_t type;
    uint8_t ssid[32];
    uint16_t channel;
    uint32_t attributes;
    uint32_t first_time;
    uint32_t last_time;
};

// Common payload for management frames talking back
typedef struct mn_common_payload {
    uint64_t sequence_id __attribute__ ((packed));
};

// Accept packet that we send back to the AP to tell it that it is accepted into
// the network, and to tell it what its station ID now is
typedef struct mn_configure_packet {
    // Reply with the sequence ID and their assigned station number
    uint64_t sequence_id __attribute__ ((packed));
    uint16_t station_id __attribute__ ((packed));

    // SSID
    uint8_t ssid[32] __attribute__ ((packed));
    
    // BSSID allocation
    uint8_t bssid0[6] __attribute__ ((packed));
    uint8_t bssid1[6] __attribute__ ((packed));
    uint8_t bssid2[6] __attribute__ ((packed));
    uint8_t bssid3[6] __attribute__ ((packed));

    // Channel assignment
    uint16_t ifchannel0 __attribute__ ((packed));
    uint16_t ifchannel1 __attribute__ ((packed));
    uint16_t ifchannel2 __attribute__ ((packed));
    uint16_t ifchannel3 __attribute__ ((packed));

    // Interface controls
    uint32_t cf_bitmap0 __attribute__ ((packed));
    uint32_t cf_bitmap1 __attribute__ ((packed));
    uint32_t cf_bitmap2 __attribute__ ((packed));
    uint32_t cf_bitmap3 __attribute__ ((packed));
};

// Activate an AP
typedef struct mn_activate_packet {
    // Not much to this...
    uint64_t sequence_id __attribute__ ((packed));
};

// Ban a user from the access point
typedef struct mn_banuser_packet {
    uint64_t sequence_id __attribute__ ((packed));
    // MAC to ban
    uint8_t banmac[6];
    // When we banned and when the ban expires
    uint32_t ban_time;
    uint32_t expire_time;
};

// Push a new MAC address into an interface in an attept to avoid an attack
typedef struct mn_bssid_packet {
    uint64_t sequence_id __attribute__ ((packed));
    uint8_t if_index __attribute__ ((packed));
    uint8_t bssid[6] __attribute__ ((packed));
};

// Smite an AP
typedef struct mn_smite_packet {
    uint64_t sequence_id __attribute__ ((packed));
    uint8_t if_index __attribute__ ((packed));
    uint8_t bssid[6] __attribute__ ((packed));
    uint16_t channel __attribute__ ((packed));
};
   
#endif

