/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

/* According to POSIX 1003.1-2001 */
#include <sys/select.h>

/* According to earlier standards */
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>

#include "messagebus.h"
#include "tcpserver.h"
#include "tcpsslserver.h"
#include "tcpsslapserver.h"
#include "tcpmonitorserver.h"
#include "timetracker.h"
#include "configfile.h"

// Global entities
MessageBus mesbus;
TcpSSLApServer mastersslsrv;
Timetracker timetracker;

int main(int argc, char *argv[]) {
    StdoutMessageClient stdclient;
    TcpMonitorMessageClient monmesgclient;
    fd_set master_readset, master_writeset;
    int max_fd = 0;
    int ret = 0;
    ConfigFile *conf;
    char errstr[1024];
    TcpMonitorServer monsrv;

    // Ignore some signals
    signal(SIGPIPE, SIG_IGN);
    
    // Register the standard output with the message bus
    mesbus.RegisterClient(&stdclient, MSGFLAG_ALL);

    // Register the message bus with the server code
    mastersslsrv.RegisterMessageBus(&mesbus);
    mastersslsrv.RegisterTimetracker(&timetracker);

    monsrv.RegisterMessageBus(&mesbus);
    monsrv.RegisterTimetracker(&timetracker);
    monsrv.RegisterApServer(&mastersslsrv);
   
    conf = new ConfigFile;
    if (conf->ParseConfig("hope_central.conf") < 0) {
        mesbus.InjectMessage("Failed to parse config file", 
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }
    
    // Set the server cert
    if (mastersslsrv.SetCTXCert(conf->FetchOpt("certpem").c_str(),
                                conf->FetchOpt("certpw").c_str()) < 0) {
        exit(1);
    }

    // Set the peer verification path
    if (mastersslsrv.SetCTXVerify(conf->FetchOpt("rootpem").c_str(),
                                  conf->FetchOpt("rootpath").c_str()) < 0) {
        exit(1);
    }

    // Set the cipher list
    if (mastersslsrv.SetCTXCipherList(conf->FetchOpt("cipherlist").c_str()) < 0) {
        exit(1);
    }

    int port, maxclients;

    if (sscanf(conf->FetchOpt("tcpport").c_str(), "%d", &port) != 1) {
        mesbus.InjectMessage("Failed to parse tcpport from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }

    if (sscanf(conf->FetchOpt("maxclients").c_str(), "%d", &maxclients) != 1) {
        mesbus.InjectMessage("Failed to parse maxclients from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }

    int monport, monmaxclients;
    if (sscanf(conf->FetchOpt("montcpport").c_str(), "%d", &monport) != 1) {
        mesbus.InjectMessage("Failed to parse montcpport from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }

    if (sscanf(conf->FetchOpt("monmaxclients").c_str(), "%d", &monmaxclients) != 1) {
        mesbus.InjectMessage("Failed to parse monmaxclients from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }

    // Grab the timings
    unsigned int conf_delay, reconf_delay, dead_delay, smite_delay, smite_ignore;
    unsigned int sniff_per, rawtx_per;
    mac_addr base_bssid;

    if (sscanf(conf->FetchOpt("configure_delay").c_str(), "%u", &conf_delay) != 1) {
        mesbus.InjectMessage("Failed to parse configure_delay from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }
    if (sscanf(conf->FetchOpt("reconfigure_delay").c_str(), "%u", &reconf_delay) != 1) {
        mesbus.InjectMessage("Failed to parse reconfigure_delay from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }
    if (sscanf(conf->FetchOpt("deadidle_delay").c_str(), "%u", &dead_delay) != 1) {
        mesbus.InjectMessage("Failed to parse deadidle_delay from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }
    if (sscanf(conf->FetchOpt("smite_delay").c_str(), "%u", &smite_delay) != 1) {
        mesbus.InjectMessage("Failed to parse smite_delay from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }
    if (sscanf(conf->FetchOpt("smite_ignore").c_str(), "%u", &smite_ignore) != 1) {
        mesbus.InjectMessage("Failed to parse smite_ignore from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }
    if (sscanf(conf->FetchOpt("num_sniffers").c_str(), "%u", &sniff_per) != 1) {
        mesbus.InjectMessage("Failed to parse num_sniffers from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }
    if (sscanf(conf->FetchOpt("num_rawtx").c_str(), "%u", &rawtx_per) != 1) {
        mesbus.InjectMessage("Failed to parse num_rawtx from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }

    base_bssid = mac_addr(conf->FetchOpt("base_bssid").c_str());
    if (base_bssid.error) {
        mesbus.InjectMessage("Failed to parse base_bssid from config file",
                              MSGFLAG_ERROR | MSGFLAG_FATAL);
        exit(1);
    }

    // Push them into the server
    mastersslsrv.SetTimeLimits(conf_delay, reconf_delay, dead_delay, 
                               smite_delay, smite_ignore);
    mastersslsrv.SetInterfaceAllocation(sniff_per, rawtx_per, 
                                        base_bssid, conf->FetchOpt("base_ssid").c_str());
    

    // Set up the server and start it
    if (mastersslsrv.Setup(maxclients, port) < 0) {
        exit(1);
    }
    snprintf(errstr, 1024, "Started SSL AP coordinator on port %d max clients %d",
             port, maxclients);
    mesbus.InjectMessage(errstr, MSGFLAG_INFO);
    snprintf(errstr, 1024, "SSID: '%s'  Configure: %us  Reconfigure: %us  "
             "DeadIdle: %us", conf->FetchOpt("base_ssid").c_str(),
             conf_delay,  reconf_delay, dead_delay);
    mesbus.InjectMessage(errstr, MSGFLAG_INFO);
    snprintf(errstr, 1024, "SmiteDelay: %us  SmiteIgnore: %us", 
             smite_delay, smite_ignore);
    mesbus.InjectMessage(errstr, MSGFLAG_INFO);


    // Set up the monitor server and start it
    if (monsrv.Setup(monmaxclients, monport) < 0) {
        exit(1);
    }
    snprintf(errstr, 1024, "Started monitor server on port %d max clients %d",
             monport, monmaxclients);
    mesbus.InjectMessage(errstr, MSGFLAG_INFO);

    monmesgclient.RegisterMonitorServer(&monsrv);
    mesbus.RegisterClient(&monmesgclient, MSGFLAG_ALL);

    // Free up mem
    delete conf;
    conf = NULL;
    
    while (1) {
        struct timeval tm;

        tm.tv_sec = 0;
        tm.tv_usec = 100000;
        
        FD_ZERO(&master_readset);
        FD_ZERO(&master_writeset);

        if ((max_fd = monsrv.MergeSet(master_readset, master_writeset,
                                      max_fd, &master_readset, &master_writeset)) < 0) {
            fprintf(stderr, "FATAL: %s\n", monsrv.FetchError());
            exit(1);
        }

        if ((max_fd = mastersslsrv.MergeSet(master_readset, master_writeset,
                                      max_fd, &master_readset, &master_writeset)) < 0) {
            fprintf(stderr, "FATAL: %s\n", mastersslsrv.FetchError());
            exit(1);
        }

        if (select(max_fd + 1, &master_readset, &master_writeset, NULL, &tm) < 0) {
            fprintf(stderr, "FATAL: select() failed: %s\n", strerror(errno));
            exit(1);
        }

        if ((ret = monsrv.Poll(master_readset, master_writeset)) < 0) {
            fprintf(stderr, "FATAL: %s\n", monsrv.FetchError());
            exit(1);
        } else if (ret == 0) {
            fprintf(stderr, "WARNING: %s\n", monsrv.FetchError());
        }

        if ((ret = mastersslsrv.Poll(master_readset, master_writeset)) < 0) {
            fprintf(stderr, "FATAL: %s\n", mastersslsrv.FetchError());
            exit(1);
        } else if (ret == 0) {
            fprintf(stderr, "WARNING: %s\n", mastersslsrv.FetchError());
        }

        mastersslsrv.Tick();
        timetracker.Tick();

    }

}

