/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

/* According to POSIX 1003.1-2001 */
#include <sys/select.h>

/* According to earlier standards */
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "messagebus.h"
#include "util.h"
#include "tcpclient.h"
#include "tcpmonitorclient.h"
#include "configfile.h"

// Global infra bits the timer needs access to
MessageBus mesbus;
TcpMonitorClient monclient;
char *fstub;

typedef struct minizstat {
    unsigned int ncli;
    unsigned long long int nframe;
    unsigned long long int nerrorframe;
    // This is plenty big enough to hold the aggregate
    unsigned long long int ntxpackets; 
    unsigned long long int nrxpackets; 
    unsigned long long int ntxbytes; 
    unsigned long long int nrxbytes; 
};

int WriteStatsTimer(Timetracker::timer_event *evt, void *parm) {
    char fname[1024];
    char errstr[1024];
    FILE *sf;

    map<unsigned int, TcpMonitorClient::zone *> *zmap =
        monclient.FetchZoneMapPtr();
    map<unsigned int, TcpMonitorClient::interface *> *ifmap =
        monclient.FetchIfMapPtr();
    map<unsigned int, TcpMonitorClient::ap *> *apmap =
        monclient.FetchApMapPtr();
    map<unsigned int, minizstat> zstatmap;
    map<unsigned int, unsigned int> ap2zonemap;

    // Map APs to zones
    for (map<unsigned int, TcpMonitorClient::ap *>::iterator apitr = 
         apmap->begin(); apitr != apmap->end(); ++apitr) {
        ap2zonemap[apitr->first] = apitr->second->zoneid;
    }

    // Map interfaces to minizonestats
    for (map<unsigned int, TcpMonitorClient::interface *>::iterator ifitr =
         ifmap->begin(); ifitr != ifmap->end(); ++ifitr) {
        unsigned int sid = ifitr->second->stationid;
        unsigned int zid = ap2zonemap[sid];

        minizstat n;
        if (zstatmap.find(zid) == zstatmap.end()) {
            memset(&n, 0, sizeof(minizstat));
            zstatmap[zid] = n;
        } else {
            n = zstatmap[zid];
        }

        n.ncli += ifitr->second->num_clients;
        n.nframe += ifitr->second->framecount;
        n.nerrorframe += ifitr->second->errorcount;

        n.ntxpackets += ifitr->second->tx_packetcount;
        n.nrxpackets += ifitr->second->rx_packetcount;
        n.ntxbytes += ifitr->second->tx_bytecount;
        n.nrxbytes += ifitr->second->rx_bytecount;

        zstatmap[zid] = n;
    }

    for (map<unsigned int, TcpMonitorClient::zone *>::iterator zitr = 
         zmap->begin(); zitr != zmap->end(); ++zitr) {
        snprintf(fname, 1024, "%s.zone.%d", fstub, zitr->first);
        if ((sf = fopen(fname, "a")) == NULL) {
            snprintf(errstr, 1024, "MonitorStatsClient failed to open file %s for "
                     "appending of zone stats", fname);
            mesbus.InjectMessage(errstr, MSGFLAG_FATAL);
            exit(1);
        }

        minizstat n;
        if (zstatmap.find(zitr->first) == zstatmap.end()) {
            memset(&n, 0, sizeof(minizstat));
            zstatmap[zitr->first] = n;
        } else {
            n = zstatmap[zitr->first];
        }

        // File contains:
        // TIMESTAMP ZID DIRTY NSNIFF NRAW NAP NNODES NIF 
        //                              NALERT NATK NCLI NFRAME NERROR NTXP NRXP NTXB NRXB
        fprintf(sf, "%d %d %d %d %d %d %d %d %d %d %d %llu %llu %llu %llu %llu %llu\n",
                time(0),
                zitr->second->id, zitr->second->dirty,
                zitr->second->num_sniffers, zitr->second->num_rawtx,
                zitr->second->num_ap, zitr->second->num_nodes,
                zitr->second->num_interfaces, zitr->second->num_alerts, 
                zitr->second->num_attacks,
                n.ncli, n.nframe, n.nerrorframe,
                n.ntxpackets, n.nrxpackets, n.ntxbytes, n.nrxbytes);

        fflush(sf);
        fclose(sf);
    }

    return 1;
}

int main(int argc, char *argv[]) {
    // Infrastructure bits
    StdoutMessageClient stdclient;
    fd_set master_readset, master_writeset;
    ConfigFile *conf;
    Timetracker timetracker;

    // Other info
    int max_fd = 0;
    int ret = 0;

    mesbus.RegisterClient(&stdclient, MSGFLAG_ALL);

    monclient.RegisterMessageBus(&mesbus);
    monclient.RegisterTimetracker(&timetracker);
    
    conf = new ConfigFile;
    if (conf->ParseConfig("/usr/local/hope/hope_monitor.conf") < 0) {
        fprintf(stderr, "FATAL: Failed to parse config file.\n");
        exit(1);
    }

    // Open the config files
    if (conf->FetchOpt("statsnamestub") == "") {
        fprintf(stderr, "FATAL: Failed to read statsnamestub from config file.\n");
        exit(1);
    }

    fstub = strdup(conf->FetchOpt("statsnamestub").c_str());

    string hostname;
    int port;

    if (ConfigFile::ParseHostLine(conf->FetchOpt("monitor_server"), &hostname, 
                                  &port) < 0) {
        fprintf(stderr, "FATAL: Invalid monitor_server host line '%s'\n", 
                conf->FetchOpt("monitor_server").c_str());
        exit(1);
    }
  
    // Set up the server and start it
    for (unsigned int x = 0; x < 10; x++) {
        if (monclient.Connect(hostname.c_str(), port) >= 0) 
            break;
        sleep(1);
    }

    if (monclient.Valid() == 0) {
        fprintf(stderr, "FATAL: Didn't connect to the monitor server in "
                "10 seconds.\n");
        exit(1);
    }

    // Free up the parser memory
    delete conf;
    conf = NULL;
 
    // Register the recurring event
    timetracker.RegisterTimer(SERVER_TIMESLICES_SEC * 5,
                              NULL, 1, &WriteStatsTimer, 
                              NULL);
    while (1) {
        struct timeval tv;
        tv.tv_sec = 0;
        tv.tv_usec = 100000;

        FD_ZERO(&master_readset);
        FD_ZERO(&master_writeset);

        // Get the sets from the central coordinator
        if ((max_fd = monclient.MergeSet(master_readset, master_writeset,
                                         max_fd, &master_readset, &master_writeset)) < 0) {
            mesbus.InjectMessage(monclient.FetchError(), MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - mergeset failed\n");
            exit(1);
        }

        if (select(max_fd + 1, &master_readset, &master_writeset, NULL, &tv) < 0) {
            mesbus.InjectMessage("select() failed", MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - select failed\n");
            exit(1);
        }

        if ((ret = monclient.Poll(master_readset, master_writeset)) < 0) {
            mesbus.InjectMessage(monclient.FetchError(), MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - monclient poll failed\n");
            exit(1);
        } 

        if (monclient.Valid() == 0) {
            printf("debug - monitor client failed\n");
            exit(1);
        }

        timetracker.Tick();

    }

}

