/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

/* According to POSIX 1003.1-2001 */
#include <sys/select.h>

/* According to earlier standards */
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "messagebus.h"
#include "util.h"
#include "tcpclient.h"
#include "tcpmonitorclient.h"
#include "configfile.h"

// Global message bus
MessageBus mesbus;

int main(int argc, char *argv[]) {
    // Infrastructure bits
    StdoutMessageClient stdclient;
    TcpMonitorClient monclient;
    fd_set master_readset, master_writeset;
    ConfigFile *conf;
    Timetracker timetracker;

    // Other info
    int max_fd = 0;
    int ret = 0;
    char errstr[1024];

    mesbus.RegisterClient(&stdclient, MSGFLAG_ALL);

    monclient.RegisterMessageBus(&mesbus);
    monclient.RegisterTimetracker(&timetracker);
    
    conf = new ConfigFile;
    if (conf->ParseConfig("/usr/local/hope/hope_monitor.conf") < 0) {
        fprintf(stderr, "FATAL: Failed to parse config file.\n");
        exit(1);
    }

    string hostname;
    int port;

    if (ConfigFile::ParseHostLine(conf->FetchOpt("monitor_server"), &hostname, 
                                  &port) < 0) {
        fprintf(stderr, "FATAL: Invalid monitor_server host line '%s'\n", 
                conf->FetchOpt("monitor_server").c_str());
        exit(1);
    }
  
    // Set up the server and start it
    for (unsigned int x = 0; x < 10; x++) {
        if (monclient.Connect(hostname.c_str(), port) >= 0) 
            break;
        sleep(1);
    }

    if (monclient.Valid() == 0) {
        fprintf(stderr, "FATAL: Didn't connect to the monitor server in "
                "10 seconds.\n");
        exit(1);
    }

    // Free up the parser memory
    delete conf;
    conf = NULL;
   
    while (1) {
        struct timeval tv;
        tv.tv_sec = 0;
        tv.tv_usec = 100000;

        FD_ZERO(&master_readset);
        FD_ZERO(&master_writeset);

        // Get the sets from the central coordinator
        if ((max_fd = monclient.MergeSet(master_readset, master_writeset,
                                         max_fd, &master_readset, &master_writeset)) < 0) {
            mesbus.InjectMessage(monclient.FetchError(), MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - mergeset failed\n");
            exit(1);
        }

        if (select(max_fd + 1, &master_readset, &master_writeset, NULL, &tv) < 0) {
            mesbus.InjectMessage("select() failed", MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - select failed\n");
            exit(1);
        }

        if ((ret = monclient.Poll(master_readset, master_writeset)) < 0) {
            mesbus.InjectMessage(monclient.FetchError(), MSGFLAG_ERROR | MSGFLAG_FATAL);
            printf("debug - monclient poll failed\n");
            exit(1);
        } 

        if (monclient.Valid() == 0) {
            printf("debug - monitor client failed\n");
            exit(1);
        }

        timetracker.Tick();

    }

}

