/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <time.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <errno.h>
#include <netpacket/packet.h>
#include <sys/ioctl.h>
#include <linux/wireless.h>
#include <fcntl.h>
#include <dirent.h>

#include "util.h"
#include "ifcontrol.h"
#include "iwcontrol.h"
#include "hostapcontrol.h"

int HostapStatFile::ParseStats(const char *in_fname) {
    FILE *configf;
    char confline[8192];

    stat_map.clear();
    
    if ((configf = fopen(in_fname, "r")) == NULL) {
        snprintf(errstr, 1024, "Reading stats file '%s': %s", 
                 in_fname, strerror(errno));
        return -1;
    }

    while (!feof(configf)) {
        fgets(confline, 8192, configf);

        if (feof(configf)) break;

        string parsestr = StrStrip(confline);
        vector<string> ldat;

        if (parsestr.length() == 0)
            continue;

        ldat = StrTokenize(parsestr, "=");

        // Skip errors
        if (ldat.size() != 2) {
            continue;
        }

        unsigned long int val;
        if (sscanf(ldat[1].c_str(), "%lu", &val) != 1) {
            snprintf(errstr, 1024, "Reading stats file '%s': Malformed value '%s'",
                     in_fname, ldat[1].c_str());
            fclose(configf);
            return -1;
        }

        stat_map[StrLower(ldat[0])] = val;
    }

    fclose(configf);

    return 1;
}

unsigned long int HostapStatFile::FetchStat(string in_key) {
    map<string, unsigned long int>::iterator sitr = stat_map.find(StrLower(in_key));

    if (sitr != stat_map.end())
        return sitr->second;

    return 0;
}

int HostAPFetchStations(const char *in_hostdir, char *errstr,
                        map<mac_addr, hostap_station_rec> *ret_map) {
    DIR *hdir;
    struct dirent *cf;
    char sname[512];
    HostapStatFile sfile;

    ret_map->clear();

    if ((hdir = opendir(in_hostdir)) == NULL) {
        snprintf(errstr, 1024, "ParseHostAPStations failed to open dir '%s': %s",
                 in_hostdir, strerror(errno));
        return -1;
    }

    // Crawl the files in it
    while ((cf = readdir(hdir)) != NULL) {
        // The cheapest way to test if the filename is what we want is to just
        // try to make it a mac address.
        mac_addr sta_mac = cf->d_name;
        
        if (sta_mac.error)
            continue;

        // Parse it like a stat file
        snprintf(sname, 512, "%s/%s", in_hostdir, cf->d_name);
        if (sfile.ParseStats(sname) < 0) {
            snprintf(errstr, 1024, sfile.FetchError());
            closedir(hdir);
            return -1;
        }

        hostap_station_rec srec;
        srec.sta_mac = sta_mac;
        srec.users = sfile.FetchStat("users");
        srec.aid = sfile.FetchStat("aid");
        srec.rxpackets = sfile.FetchStat("rx_packets");
        srec.txpackets = sfile.FetchStat("tx_packets");
        srec.rxbytes = sfile.FetchStat("rx_bytes");
        srec.txbytes = sfile.FetchStat("tx_bytes");

        (*ret_map)[sta_mac] = srec;

    }

    closedir(hdir);

    return 1;
}

int HostAPCreateRawSock(const char *in_if, char *errstr) {
    struct sockaddr_ll addr;
    struct ifreq req;
    int sock;
    char alt_if[32];

    snprintf(alt_if, 32, "%sap", in_if);
    
    // Set it to mode master
    if (Iwconfig_Set_Mode(in_if, errstr, LINUX_WLEXT_MASTER) < 0) {
        return -1;
    }

    // Set the driver to be in hostapd mode
    if (Iwconfig_Set_IntPriv(in_if, "hostapd", 1, 0, errstr) < 0) {
        return -1;
    }

    // Turn off beaconing
    if (Iwconfig_Set_IntPriv(in_if, "beacon_int", 0, 0, errstr) < 0) {
        return -1;
    }

    // Bring the interface up
    if (Ifconfig_Delta_Flags(in_if, errstr, (IFF_UP | IFF_PROMISC)) < 0) {
        return -1;
    }

    // Bring the alternate interface up
    if (Ifconfig_Delta_Flags(alt_if, errstr, (IFF_UP | IFF_PROMISC)) < 0) {
        return -1;
    }

    // build a raw packet socket
    if ((sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL))) < 0) {
        snprintf(errstr, 1024, "HostAPCreateRawSock failed to create PF_PACKET "
                 "raw socket: %s", strerror(errno));
        return -1;
    }

    // Get the interface index
    memset(&req, 0, sizeof(struct ifreq));
    strncpy(req.ifr_name, alt_if, sizeof(req.ifr_name));

    if (ioctl(sock, SIOCGIFINDEX, &req) < 0) {
        snprintf(errstr, 1024, "HostAPCreateRawSock failed to fetch interface "
                 "index for '%s': %s", in_if, strerror(errno));
        close(sock);
        return -1;
    }

    if (Ifconfig_Set_MTU(alt_if, errstr, HOSTAP_RAWTX_MTU) < 0) {
        return -1;
    }
    
    // Bind the socket to the interface
    memset(&addr, 0, sizeof(struct sockaddr_ll));
    addr.sll_ifindex = req.ifr_ifindex;
    addr.sll_family = AF_PACKET;

    if (bind(sock, (struct sockaddr *) &addr, sizeof(struct sockaddr_ll)) < 0) {
        snprintf(errstr, 1024, "HostAPCreateRawSock failed to bind interface '%s': %s",
                 in_if, strerror(errno));
        close(sock);
        return -1;
    }

    if (Iwconfig_Set_SSID(in_if, errstr, " ") < 0) {
        return -1;
    }

    return sock;
}

int HostAPWriteFrame(int in_sock, uint8_t *in_xmit, int in_len, char *errstr) {
    int wlen;

    if ((wlen = send(in_sock, in_xmit, in_len, 0)) <= 0) {
        snprintf(errstr, 1024, "HostAPWriteFrame failed to send() data: %s",
                 strerror(errno));
        return -1;
    }

    return wlen;
}

