/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include "util.h"
#include "macaddr.h"
#include "tcpkismetclient.h"

TcpKismetClient::TcpKismetClient() {
    // Fill in the default protocol stuff
    protocol_default_map["INFO"] = "networks,packets,dropped,rate";
    protocol_default_map["ALERT"] = "header,bssid,source,dest,other,channel";
    protocol_default_map["KISMET"] = "version,starttime,servername,timestamp,"
        "channelhop,newversion";
    // Limited version of the network protocol
    protocol_default_map["NETWORK"] = "bssid,type,ssid,wep,channel,firsttime,lasttime";

    protocol_map["INFO"] = 1;
    protocol_map["ALERT"] = 1;
    protocol_map["KISMET"] = 1;
    protocol_map["NETWORK"] = 1;
    
    srvtime = 0;
    num_networks = num_packets = num_dropped = 0;
    old_num_networks = old_num_packets = old_num_dropped = 0;
    packet_rate = 0;

    networklist_dirty = 0;
    alertlist_dirty = 0;
}

TcpKismetClient::~TcpKismetClient() {

}

int TcpKismetClient::Connect(const char *in_remotehost, short int in_port) {
    // Shunt to the normal connect stuff
    // printf("debug - kismet connect\n");
    if (TcpClient::Connect(in_remotehost, in_port) < 0)
        return -1;

    // printf("debug - kismet connected %d\n", client_fd);
    // Turn on the protocols in our map, if we have any.  This is for reconnecting
    // to a server and preserving the requested protocols
    for (map<string, int>::iterator prot = protocol_map.begin();
         prot != protocol_map.end(); ++prot) {
        EnableProtocol(prot->first);
    }

    return 1;
}

void TcpKismetClient::EnableProtocol(string in_protocol) {
    // printf("debug - enable protocol %s\n", in_protocol.c_str());
    string fields;
    char data[1024];

    if (protocol_default_map.find(in_protocol) != protocol_default_map.end())
        fields = protocol_default_map[in_protocol];
    else
        fields = "*";

    // We don't care about ACK or errors, because we suck
    snprintf(data, 1024, "!0 ENABLE %s %s\n", in_protocol.c_str(), fields.c_str());

    protocol_map[in_protocol] = 1;
    
    if (cl_valid == 0)
        return;
    
    // Not much we can do if this fails...
    if (QueueData(data) < 0)
        KillConnection();
}

int TcpKismetClient::FetchNetworks(vector<TcpKismetClient::network_rec *> *ret_vec) {
    (*ret_vec) = network_vec;
    networklist_dirty = 0;
    return network_vec.size();
}

int TcpKismetClient::FetchAlerts(vector<TcpKismetClient::alert_rec *> *al_vec) {
    (*al_vec) = alert_vec;
    alertlist_dirty = 0;
    return alert_vec.size();
}

int TcpKismetClient::ParseData() {
    char *srvdata;
    int readlen = 0;

    // Try to fetch a complete string from the server...  We should never
    // get called if we have a 0 length ring, so we can do an unshielded
    // malloc
    readlen = read_buf->FetchLen();
    srvdata = new char[readlen + 1];

    // Fetch it into the char*
    read_buf->FetchPtr((uint8_t *) srvdata, readlen, &readlen);
    srvdata[readlen] = '\0';

    // Vectorize it on a \n
    vector<string> inptok = StrTokenize(srvdata, "\n");

    // Nuke our raw buffer since we don't need it anymore
    delete[] srvdata;

    // Bail on no useful data
    if (inptok.size() <= 1) {
        return 0;
    }

    // Go through all but the last one, which probably isn't a full line
    // Process and mark each one read
    for (unsigned int it = 0; it < (inptok.size() - 1); it++) {
        read_buf->MarkRead(inptok[it].length() + 1);

        char protoh[65];
        if (sscanf(inptok[it].c_str(), "%64[^:]", protoh) < 1) {
            snprintf(errstr, 1024, "Unable to detect protocol on Kismet line '%s'",
                     inptok[it].c_str());
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            continue;
        }

        // Dispatch it out (and increment past the proto header)
        if (!strncmp(protoh, "*TIME", 64)) {
            ParseTimeProto(inptok[it].c_str() + strlen(protoh) + 2);
        } else if (!strncmp(protoh, "*KISMET", 64)) {
            ParseKismetProto(inptok[it].c_str() + strlen(protoh) + 2);
        } else if (!strncmp(protoh, "*PROTOCOLS", 64)) {
            ParseProtocolsProto(inptok[it].c_str() + strlen(protoh) + 2);
        } else if (!strncmp(protoh, "*INFO", 64)) {
            ParseInfoProto(inptok[it].c_str() + strlen(protoh) + 2);
        } else if (!strncmp(protoh, "*ALERT", 64)) {
            ParseAlertProto(inptok[it].c_str() + strlen(protoh) + 2);
        } else if (!strncmp(protoh, "*NETWORK", 64)) {
            ParseNetworkProto(inptok[it].c_str() + strlen(protoh) + 2);
        } else {
            snprintf(errstr, 1024, "Unknown protocol '%s' on Kismet line '%s'",
                     protoh, inptok[it].c_str());
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            continue;
        }
        
    }

    return 1;
}

void TcpKismetClient::ParseTimeProto(const char *in_line) {
    if (sscanf(in_line, "%d", (int *) &srvtime) < 1) {
        snprintf(errstr, 1024, "Corrupt TIME protocol line '%s' from Kismet",
                 in_line);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }
}

void TcpKismetClient::ParseKismetProto(const char *in_line) {
    int junkmajor, junkminor, junktiny;

    if (sscanf(in_line, "%d.%d.%d %d \001%32[^\001]\001 %24s %d %24[^.].%24[^.].%24s",
               &junkmajor, &junkminor, &junktiny, 
               (int *) &start_time, servername, 
               build, &channel_hop, major, minor, tiny) < 7) {
        snprintf(errstr, 1024, "Corrupt KISMET protocol line '%s' from Kismet",
                 in_line);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }

    snprintf(errstr, 1024, "TcpKismetClient connected to Kismet server '%s' version "
             "%s.%s.%s build %s", servername, major, minor, tiny, build);
    msgbus->InjectMessage(errstr, MSGFLAG_INFO);
}

void TcpKismetClient::ParseProtocolsProto(const char *in_line) {
    // We don't actually even do anything with this.  We're bad.
}

void TcpKismetClient::ParseInfoProto(const char *in_line) {
    old_num_networks = num_networks;
    old_num_packets = num_packets;
    old_num_dropped = num_dropped;

    if (sscanf(in_line, "%d %d %d %d", 
               &num_networks, &num_packets, &num_dropped, &packet_rate) != 4) {
        snprintf(errstr, 1024, "TcpKismetClient Corrupt INFO protocol line '%s'",
                 in_line);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }

    /*
     * printf("debug - got INFO frame, %d nets %d packets %d dropped %d rate\n",
           num_networks, num_packets, num_dropped, packet_rate);
           */
}

void TcpKismetClient::ParseNetworkProto(const char *in_line) {
    char pbssid[MAC_STR_LEN];
    int ptype;
    char pssid[64];
    int pfirst;
    int plast;
    int pwep;
    int pchannel;
    network_rec *nrec;

    if (sscanf(in_line, "%17s %d \001%32[^\001]\001 %d %d %d %d",
               pbssid, &ptype, pssid, &pwep, &pchannel, &pfirst, &plast) != 7) {
        snprintf(errstr, 1024, "TcpKismetClient Corrupt NETWORK "
                 "protocol line '%s'", in_line);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }

    macmap<TcpKismetClient::network_rec *>::iterator nitr =
        network_map.find(pbssid);
    if (nitr != network_map.end()) {
        nrec = *(nitr->second);
    } else {
        nrec = new network_rec;
        network_map.insert(pbssid, nrec);
        network_vec.push_back(nrec);
    }

    nrec->bssid_mac = pbssid;
    nrec->type = (kis_network_type) ptype;
    nrec->ssid = pssid;
    nrec->channel = pchannel;
    nrec->first_time = (time_t) pfirst;
    nrec->last_time = (time_t) plast;

    nrec->attributes = 0;
    if (pwep)
        nrec->attributes |= KISNETATTRIB_WEP;

    /* printf("debug - got kismet network %s type %d ssid %s chan %d first %d last %d attrib %d\n",
           nrec->bssid_mac.Mac2String().c_str(), nrec->type, nrec->ssid.c_str(), nrec->channel,
           nrec->first_time, nrec->last_time, nrec->attributes);*/
    
    networklist_dirty = 1;
}

void TcpKismetClient::ParseAlertProto(const char *in_line) {
    char proto_header[32];
    char proto_bssid[MAC_STR_LEN];
    char proto_source[MAC_STR_LEN];
    char proto_dest[MAC_STR_LEN];
    char proto_other[MAC_STR_LEN];
    mac_addr bssid, source, dest, other;
    int proto_channel;
    alert_rec *alr = NULL;

    if (sscanf(in_line, "%32s %17s %17s %17s %17s %d",
               proto_header, proto_bssid, proto_source, 
               proto_dest, proto_other, &proto_channel) != 6) {
        snprintf(errstr, 1024, "TcpKismetClient Corrupt ALERT protocol line '%s'",
                 in_line);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }

    bssid = proto_bssid;
    if (bssid.error) { 
        snprintf(errstr, 1024, "TcpKismetClient corrupt BSSID mac in alert '%s'",
                 in_line);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }
    
    source = proto_source;
    if (source.error) { 
        snprintf(errstr, 1024, "TcpKismetClient corrupt SOURCE mac in alert '%s'",
                 in_line);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }

    dest = proto_dest;
    if (dest.error) { 
        snprintf(errstr, 1024, "TcpKismetClient corrupt DEST mac in alert '%s'",
                 in_line);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }

    other = proto_other;
    if (other.error) { 
        snprintf(errstr, 1024, "TcpKismetClient corrupt OTHER mac in alert '%s'",
                 in_line);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return;
    }

    if (!strncmp(proto_header, "BSSTIMESTAMP", 32)) {
        alr = new TcpKismetClient::alert_rec;
        alr->type = kis_alrt_bsstimestamp;
        alr->bssid_mac = bssid;

        // printf("debug - BSSTIMESTAMP alert on bssid %s\n", bssid.Mac2String().c_str());
    } else if (!strncmp(proto_header, "DEAUTHFLOOD", 32)) {
        alr = new TcpKismetClient::alert_rec;
        alr->type = kis_alrt_deauthflood;
        alr->bssid_mac = bssid;

        // printf("debug - DEAUTHFLOOD alert on bssid %s\n", bssid.Mac2String().c_str());
    } else if (!strncmp(proto_header, "BCASTDISCON", 32)) {
        alr = new TcpKismetClient::alert_rec;
        alr->type = kis_alrt_bcastdiscon;
        alr->bssid_mac = bssid;

        // printf("debug - BCASTDISCON alert on bssid %s\n", bssid.Mac2String().c_str());
    } else if (!strncmp(proto_header, "NULLPROBERESP", 32)) {
        alr = new TcpKismetClient::alert_rec;
        alr->type = kis_alrt_nullproberesp;
        alr->bssid_mac = bssid;
        alr->dest_mac = dest;
        alr->source_mac = source;

        // printf("debug - NULLPROBERESP alert on bssid %s source %s dest %s\n",
               // bssid.Mac2String().c_str(), source.Mac2String().c_str(), dest.Mac2String().c_str());
    } else {
        // printf("debug - alert we don't handle %s\n", proto_header);
    }

    if (alr != NULL) {
        alr->channel = proto_channel;
        alr->timestamp = time(0);
        // Hardcoded max, this is probably bad
        if (alert_vec.size() >= 20) {
            delete alert_vec[0];
            alert_vec.erase(alert_vec.begin());
        }
        alert_vec.push_back(alr);

        alertlist_dirty = 1;
    }
}

int TcpKismetClient::QueueData(const char *in_line) {
    if (write_buf->InsertDummy(strlen(in_line)) == 0) {
        snprintf(errstr, 1024, "TcpKismetClient - no room in write ring buffer "
                 "to insert a queued line of len %d", strlen(in_line));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    write_buf->InsertData((uint8_t *) in_line, strlen(in_line));

    return 1;
}

