/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __TCPKISMETCLIENT_H__
#define __TCPKISMETCLIENT_H__

#include "config.h"

#include <string>
#include <vector>

#include "tcpclient.h"
#include "macaddr.h"

#define KISNETATTRIB_NONE       0
#define KISNETATTRIB_WEP        1

class TcpKismetClient : public TcpClient {
public:
    enum kis_network_type {
        network_ap,
        network_adhoc,
        network_probe,
        network_turbocell,
        network_data,
        network_remove
    };

    typedef struct network_rec {
        mac_addr bssid_mac;
        kis_network_type type;
        string ssid;
        time_t first_time;
        time_t last_time;
        int channel;
        uint32_t attributes;

        // For the tracker to fill in
        time_t tracker_time;
        void *ext;
    };

    enum kis_alert_type {
        kis_alrt_unknown, 
        kis_alrt_bsstimestamp,
        kis_alrt_deauthflood,
        kis_alrt_bcastdiscon,
        kis_alrt_nullproberesp
    };

    typedef struct alert_rec {
        kis_alert_type type;
        mac_addr bssid_mac;
        mac_addr source_mac;
        mac_addr dest_mac;
        mac_addr other_mac;
        int channel;
        time_t timestamp;

        // For the tracker to fill in
        time_t tracker_time;
        void *ext;
    };

    TcpKismetClient();
    virtual ~TcpKismetClient();

    // Override some of the connect stuff
    virtual int Connect(const char *in_remotehost, short int in_port);

    // Turn on a protocol we know how to handle
    virtual void EnableProtocol(string in_proto);

    // Fetch list of networks
    virtual int FetchNetworkDirty() {
        return networklist_dirty;
    }
    virtual int FetchNetworks(vector<TcpKismetClient::network_rec *> *ret_netvec);
    
    // Fetch list of alerts
    virtual int FetchAlertDirty() {
        return alertlist_dirty;
    }
    virtual int FetchAlerts(vector<TcpKismetClient::alert_rec *> *ret_alvec);

protected:
    virtual int ParseData();

    // Various internal kismet protocol actions to be dispatched by
    // ParseData().  These process the lines and fill in the local 
    // structs for all the fetching
    virtual void ParseTimeProto(const char *in_line);
    virtual void ParseKismetProto(const char *in_line);
    virtual void ParseProtocolsProto(const char *in_line);
    virtual void ParseInfoProto(const char *in_line);
    virtual void ParseNetworkProto(const char *in_line);
    virtual void ParseAlertProto(const char *in_line);

    // queue data
    virtual int QueueData(const char *in_line);
    
    // Fields we handle in the protocols now
    map<string, string> protocol_default_map;
    map<string, int> protocol_map;

    // States from the Kismet server
    time_t srvtime;
    time_t start_time;

    char servername[32];
    char major[24], minor[24], tiny[24], build[24];

    int channel_hop;

    int num_networks, num_packets, num_dropped;
    int old_num_networks, old_num_packets, old_num_dropped;

    int packet_rate;

    int networklist_dirty;
    int alertlist_dirty;

    macmap<TcpKismetClient::network_rec *> network_map;
    vector<TcpKismetClient::network_rec *> network_vec;

    vector<TcpKismetClient::alert_rec *> alert_vec;
};

#endif

