/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include "tcpmonitorserver.h"
#include "endian_magic.h"

int MonitorPushTimer(Timetracker::timer_event *evt, void *parm) {
    TcpMonitorServer *ms = (TcpMonitorServer *) parm;

    ms->PushTimer();
    
    return 1;
}

TcpMonitorServer::TcpMonitorServer() {
    apsslserver = NULL;
    timer_id = 0;
}

TcpMonitorServer::~TcpMonitorServer() {
    // Null
}

void TcpMonitorServer::RegisterApServer(TcpSSLApServer *in_apserver) {
    apsslserver = in_apserver;
}

int TcpMonitorServer::ParseData(int in_fd) {
    RingBuffer *rbuf = read_buf_map[in_fd];

    rbuf->MarkRead(rbuf->FetchLen());
    
    return 1;
}

int TcpMonitorServer::Setup(unsigned int in_max_clients, short int in_port) {
    if (timer_id == 0) {
        timer_id = timetracker->RegisterTimer(SERVER_TIMESLICES_SEC,
                                              NULL, 1, &MonitorPushTimer, 
                                              (void *) this);
    }

    return (TcpServer::Setup(in_max_clients, in_port));
}

void TcpMonitorServer::KillConnection(int in_fd) {
    map<int, TcpMonitorServer::client_info *>::iterator mitr = monitor_recs.find(in_fd);
    if (mitr != monitor_recs.end()) {
        delete mitr->second;
        monitor_recs.erase(mitr);
    }

    TcpServer::KillConnection(in_fd);

    snprintf(errstr, 1024, "Terminated monitor connection on fd %d", in_fd);
    msgbus->InjectMessage(errstr, MSGFLAG_INFO);

}

int TcpMonitorServer::Validate(int in_fd) {
    if (ValidateIPFilter(in_fd) < 0)
        return -1;

    monitor_recs[in_fd] = new TcpMonitorServer::client_info;
    monitor_recs[in_fd]->dead = 0;
    monitor_recs[in_fd]->pristine = 1;
    monitor_recs[in_fd]->seqid = rand();

    snprintf(errstr, 1024, "Accepted new monitor connection on fd %d", in_fd);
    msgbus->InjectMessage(errstr, MSGFLAG_INFO);

    return 1;
}

void TcpMonitorServer::PushTimer() {
    // Push everything that is newer than, or not present in, each client
    // record.
    for (map<int, TcpMonitorServer::client_info *>::iterator mitr = monitor_recs.begin();
         mitr != monitor_recs.end(); ++mitr) {
        client_info *cl = mitr->second;
        time_t curtime = time(0);

        if (cl->dead)
            continue;

        if (cl->pristine) {
            PushInfoRecord(mitr->first);
            cl->pristine = 0;
        }

        // Send crap to the client
        for (map<int, TcpSSLApServer::apserver_zonerec>::iterator zitr = 
             apsslserver->zone_map.begin(); zitr != apsslserver->zone_map.end();
             ++zitr) {

            // First we need to see if they need to know about the zone itself
            if (cl->zone_sent_map.find(zitr->first) == cl->zone_sent_map.end()) {
                cl->zone_sent_map[zitr->first] = curtime;

                // push and skip rest on error 
                if (PushZoneRecord(mitr->first, &(zitr->second)) < 0) {
                    cl->dead = 1;
                    break;
                }

            } else if (cl->zone_sent_map[zitr->first] < zitr->second.update_time) {
                cl->zone_sent_map[zitr->first] = curtime;
                // push and skip rest on error
                if (PushZoneRecord(mitr->first, &(zitr->second)) < 0) {
                    cl->dead = 1;
                    break;
                }
            }

            // Bail if we didn't sent a zone correctly
            if (cl->dead)
                continue;

            // Send all the interfaces in a zone, since this is a convenient index to them
            for (unsigned int x = 0; x < zitr->second.interfaces.size(); x++) {
                TcpSSLApServer::apserver_apinterface *ifc = zitr->second.interfaces[x];
                if (cl->if_sent_map.find(ifc) == cl->if_sent_map.end()) {
                    cl->if_sent_map[ifc] = curtime;
                    
                    if (PushIfRecord(mitr->first, ifc) < 0) {
                        cl->dead = 1;
                        break;
                    }
                } else if (cl->if_sent_map[ifc] < ifc->update_time) {
                    cl->if_sent_map[ifc] = curtime;
                    
                    if (PushIfRecord(mitr->first, ifc) < 0) {
                        cl->dead = 1;
                        break;
                    }
                }
            }
        }

        if (cl->dead)
            continue;

        // Crawl through all the APs in the general map
        for (map<int, TcpSSLApServer::apserver_aprec *>::iterator aitr =
             apsslserver->linked_ap_map.begin(); aitr != apsslserver->linked_ap_map.end();
             ++aitr) {
            if (cl->ap_sent_map.find(aitr->second) == cl->ap_sent_map.end()) {
                cl->ap_sent_map[aitr->second] = curtime;

                if (PushApRecord(mitr->first, aitr->second) < 0) {
                    cl->dead = 1;
                    break;
                }
            } else if (cl->ap_sent_map[aitr->second] < aitr->second->update_time) {
                cl->ap_sent_map[aitr->second] = curtime;

                if (PushApRecord(mitr->first, aitr->second) < 0) {
                    cl->dead = 1;
                    break;
                }
            }
        }
    }

    // HANDLE EXPIRING DEAD PEOPLE HERE
    for (map<int, TcpMonitorServer::client_info *>::iterator mitr = monitor_recs.begin();
         mitr != monitor_recs.end(); ++mitr) {

        if (mitr->second->dead == 0)
            continue;
        map<int, TcpMonitorServer::client_info *>::iterator ditr = mitr;
        int beg = 0;

        if (mitr != monitor_recs.begin())
            --ditr;
        else
            beg = 1;

        KillConnection(mitr->first);

        if (monitor_recs.size() == 0)
            break;

        if (beg == 1)
            mitr = monitor_recs.begin();
        else
            mitr = ditr;
    }

}

int TcpMonitorServer::PushInfoRecord(int in_fd) {
    monitor_frame mf;

    memset(&mf, 0, sizeof(monitor_frame));

    // Fill in the header
    mf.sentinel = (uint32_t) htonl(MONITOR_SENTINEL);
    mf.type = MONTYPE_SERVERINFO;
    mf.len = ntohl(sizeof(mf.u.serverinfo));

    mf.u.serverinfo.proto_ver = MONITOR_PROTOVER;
    mf.u.serverinfo.start_time = (uint32_t) htonl(time(0));
    mf.u.serverinfo.config_delay = (uint16_t) htons(apsslserver->config_delay);
    mf.u.serverinfo.reconfig_delay = (uint16_t) htons(apsslserver->reconfigure_delay);
    mf.u.serverinfo.deadidle_delay = (uint16_t) htons(apsslserver->deadidle_delay);
    mf.u.serverinfo.smite_delay = (uint16_t) htons(apsslserver->smite_delay);
    mf.u.serverinfo.smite_ignore = (uint16_t) htons(apsslserver->smite_ignore);
    mf.u.serverinfo.sniffers_per = (uint16_t) htons(apsslserver->sniffers_per_zone);
    mf.u.serverinfo.rawtx_per = (uint16_t) htons(apsslserver->rawtx_per_zone);

    // Ignore bssid and bssid_mask

    strncpy((char *) mf.u.serverinfo.base_ssid, apsslserver->base_ssid, 32);

    return QueuePacket(in_fd, &mf);
}

int TcpMonitorServer::PushZoneRecord(int in_fd, TcpSSLApServer::apserver_zonerec *in_zone) {
    monitor_frame mf;

    memset(&mf, 0, sizeof(monitor_frame));

    // Fill in the header
    mf.sentinel = (uint32_t) htonl(MONITOR_SENTINEL);
    mf.type = MONTYPE_ZONE;
    mf.len = (uint32_t) htonl(sizeof(mf.u.zone));

    mf.u.zone.id = (uint16_t) htons(in_zone->zone_id);
    mf.u.zone.dirty = (uint8_t) in_zone->zone_dirty;
    mf.u.zone.nsniffer = (uint8_t) in_zone->num_sniffer;
    mf.u.zone.nrawtx = (uint8_t) in_zone->num_rawtx;
    mf.u.zone.nap = (uint8_t) in_zone->num_ap;
    mf.u.zone.nnodes = (uint8_t) in_zone->aps.size();
    mf.u.zone.ninterfaces = (uint8_t) in_zone->interfaces.size();
    mf.u.zone.nalerts = (uint32_t) htonl(in_zone->num_alerts);
    mf.u.zone.nattacks = (uint32_t) htonl(in_zone->num_attacks);

    return QueuePacket(in_fd, &mf);
}

int TcpMonitorServer::PushApRecord(int in_fd, TcpSSLApServer::apserver_aprec *in_ap) {
    monitor_frame mf;

    memset(&mf, 0, sizeof(monitor_frame));

    // Fill in the header
    mf.sentinel = (uint32_t) htonl(MONITOR_SENTINEL);
    mf.type = MONTYPE_AP;
    mf.len = (uint32_t) htonl(sizeof(mf.u.ap));

    strncpy((char *) mf.u.ap.name, in_ap->name, 32);
    mf.u.ap.zoneid = (uint16_t) htons(in_ap->zone);
    mf.u.ap.stationid = (uint16_t) htons(in_ap->station_id);
    mf.u.ap.configured = (uint8_t) in_ap->configured;
    mf.u.ap.active = (uint8_t) in_ap->active;
    mf.u.ap.dead = (uint8_t) in_ap->dead;
    mf.u.ap.last_helo = (uint32_t) htonl(in_ap->last_helo.tv_sec);
    // Change this from being hardcoded...
    mf.u.ap.ninterfaces = (uint8_t) 4;
    mf.u.ap.loadavg_major = (uint8_t) in_ap->load_avg_major;
    mf.u.ap.loadavg_minor = (uint8_t) in_ap->load_avg_minor;

    return QueuePacket(in_fd, &mf);
}

int TcpMonitorServer::PushIfRecord(int in_fd, TcpSSLApServer::apserver_apinterface *in_if) {
    monitor_frame mf;

    memset(&mf, 0, sizeof(monitor_frame));

    // Fill in the header
    mf.sentinel = (uint32_t) htonl(MONITOR_SENTINEL);
    mf.type = MONTYPE_INTERFACE;
    mf.len = (uint32_t) htonl(sizeof(mf.u.interface));

    strncpy((char *) mf.u.interface.name, in_if->name, 32);
    mf.u.interface.stationid = (uint16_t) htons(in_if->owning_ap->station_id);
    mf.u.interface.index = (uint8_t) in_if->index;
    for (unsigned int x = 0; x < 6; x++) {
        mf.u.interface.mac[x] = in_if->mac[x];
        mf.u.interface.mac[x] = in_if->innate_mac[x];
    }
    mf.u.interface.channel = (uint16_t) htons(in_if->channel);
    mf.u.interface.cf_bitmap = (uint32_t) htonl(in_if->cf_bitmap);
    mf.u.interface.last_mac_change = (uint32_t) htonl(in_if->last_mac_change);
    mf.u.interface.nclients = (uint32_t) htonl(in_if->clientcount);
    mf.u.interface.framecount = (uint64_t) kis_hton64(in_if->framecount);
    mf.u.interface.errorcount = (uint64_t) kis_hton64(in_if->errorcount);
    mf.u.interface.txpacketcount = (uint64_t) kis_hton64(in_if->tx_packetcount);
    mf.u.interface.rxpacketcount = (uint64_t) kis_hton64(in_if->rx_packetcount);
    mf.u.interface.txbytecount = (uint64_t) kis_hton64(in_if->tx_bytecount);
    mf.u.interface.rxbytecount = (uint64_t) kis_hton64(in_if->rx_bytecount);
    
    return QueuePacket(in_fd, &mf);
}

int TcpMonitorServer::PushMessageRecord(int in_fd, const char *in_msg, int in_msgflag) {
    monitor_frame mf;

    memset(&mf, 0, sizeof(monitor_frame));

    // Fill in the header
    mf.sentinel = (uint32_t) htonl(MONITOR_SENTINEL);
    mf.type = MONTYPE_MESSAGE;
    // Length of message frame sub the part of the message we don't use
    mf.len = (uint32_t) htonl(sizeof(mf.u.message) - (1024 - strlen(in_msg)));

    mf.u.message.msgflags = (uint32_t) htonl(in_msgflag);
    mf.u.message.msglen = (uint32_t) htonl(strlen(in_msg));
    strncpy((char *) mf.u.message.msgtext, in_msg, strlen(in_msg));
    
    return QueuePacket(in_fd, &mf);
}

int TcpMonitorServer::PushMessageAll(const char *in_msg, int in_msgflag) {
    for (map<int, TcpMonitorServer::client_info *>::iterator mitr = monitor_recs.begin();
         mitr != monitor_recs.end(); ++mitr) {
        if (mitr->second->dead)
            continue;
        
        if (PushMessageRecord(mitr->first, in_msg, in_msgflag) < 0) {
            mitr->second->dead = 1;
            continue;
        }
    }

    for (map<int, TcpMonitorServer::client_info *>::iterator mitr = monitor_recs.begin();
         mitr != monitor_recs.end(); ++mitr) {

        if (mitr->second->dead == 0)
            continue;
        map<int, TcpMonitorServer::client_info *>::iterator ditr = mitr;
        int beg = 0;

        if (mitr != monitor_recs.begin())
            --ditr;
        else
            beg = 1;

        KillConnection(mitr->first);

        if (monitor_recs.size() == 0)
            break;

        if (beg == 1)
            mitr = monitor_recs.begin();
        else
            mitr = ditr;
    }

    return 1;
}

int TcpMonitorServer::QueuePacket(int in_fd, monitor_frame *in_frame) {
    if (monitor_recs[in_fd]->dead) {
        snprintf(errstr, 1024, "Attempt to write to dead client %u, this is bad.",
                 in_fd);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return 0;
    }

    // printf("debug - queuing packet type %d\n", in_frame->frame_type);
    int hostlen = ntohl(in_frame->len);
    // printf("debug - %d hostlen - %d\n", in_frame->len, hostlen);
    RingBuffer *write_buf = write_buf_map[in_fd];

    // Modify the sequence ID
    unsigned long int seq = monitor_recs[in_fd]->seqid++;
    in_frame->sequence_id = (uint64_t) kis_hton64(seq);

    // Queue it into the write buffer
    if (write_buf->InsertDummy(MONLEN_HEADER + hostlen) == 0) {
        snprintf(errstr, 1024, "TCPMONITOR no room in write ring buffer to insert "
                 "queued packet.");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }
    write_buf->InsertData((uint8_t *) in_frame, MONLEN_HEADER + hostlen);

    return MONLEN_HEADER + hostlen;
}

void TcpMonitorMessageClient::ProcessMessage(string in_msg, int in_flags) {
    // Propogate it to the server
    srv->PushMessageAll(in_msg.c_str(), in_flags);
}


