#include "tcpserver.h"

TcpServer::TcpServer() {
    // Init stuff
    sv_valid = 0;

    serv_fd = 0;
    max_fd = 0;

    msgbus = NULL;
}

TcpServer::~TcpServer() {
    // Cleanup stuff
    sv_valid = 0;
}

int TcpServer::Setup(unsigned int in_max_clients, short int in_port) {
    max_clients = in_max_clients;

    // Find local host
    if (gethostname(hostname, MAXHOSTNAMELEN) < 0) {
        snprintf(errstr, 1024, "TCP server gethostname() failed: %s",
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    port = in_port;

    // Set up socket stuff
    memset(&serv_sock, 0, sizeof(serv_sock));
    serv_sock.sin_family = AF_INET;
    serv_sock.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_sock.sin_port = htons(port);

    if ((serv_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        snprintf(errstr, 1024, "TCP server socket() failed: %s",
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Reuse the addr
    int i = 2;
    if (setsockopt(serv_fd, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) == -1) {
        snprintf(errstr, 1024, "TCP server setsockopt(REUSEADDR) failed: %s",
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Bind the socket
    if (bind(serv_fd, (struct sockaddr *) &serv_sock, sizeof(serv_sock)) < 0) {
        snprintf(errstr, 1024, "TCP server bind() failed: %s",
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Listen for connections
    if (listen(serv_fd, 5) < 0) {
        snprintf(errstr, 1024, "TCP server listen() failed: %s",
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Zero and set the FDs and maxfd
    FD_ZERO(&server_fdset);

    if (serv_fd > max_fd)
        max_fd = serv_fd;

    // We're valid
    sv_valid = 1;

    return 1;
}

int TcpServer::SetIPFilter(vector<client_ipfilter *> in_filter) {
    ipfilter_vec = in_filter;

    return 1;
}

// Merge our file descriptors into an existing set
unsigned int TcpServer::MergeSet(fd_set in_rset, fd_set in_wset,
                                 unsigned int in_max_fd,
                                 fd_set *out_rset, fd_set *out_wset) {
    unsigned int max;

    FD_ZERO(out_rset);
    FD_ZERO(out_wset);
   
    if (in_max_fd < max_fd) {
        max = max_fd;
    } else {
        max = in_max_fd;
    }

    // Set the server fd
    FD_SET(serv_fd, out_rset);
    
    for (unsigned int x = 0; x <= max; x++) {
        // Incoming read or our own clients
        if (FD_ISSET(x, &in_rset) || FD_ISSET(x, &server_fdset))
            FD_SET(x, out_rset);
        // Incoming write or any clients with a pending write ring
        if (FD_ISSET(x, &in_wset) || 
            (write_buf_map.find(x) != write_buf_map.end() &&
             write_buf_map[x]->FetchLen() > 0))
            FD_SET(x, out_wset);
    }
   
    return max;
}

void TcpServer::KillConnection(int in_fd) {
    // Nuke descriptors
    FD_CLR(in_fd, &server_fdset);

    // Nuke ringbuffers
    map<int, RingBuffer *>::iterator miter = read_buf_map.find(in_fd);
    if (miter != read_buf_map.end()) {
        delete read_buf_map[in_fd];
        read_buf_map.erase(miter);
    }

    miter = write_buf_map.find(in_fd);
    if (miter != write_buf_map.end()) {
        delete write_buf_map[in_fd];
        write_buf_map.erase(miter);
    }
   
    // Close the fd
    if (in_fd)
        close(in_fd);

}

void TcpServer::Shutdown() {
    for (map<int, RingBuffer *>::iterator miter = read_buf_map.begin();
         miter != read_buf_map.end(); miter++)
        KillConnection(miter->first);

    sv_valid = 0;

    if (serv_fd)
        close(serv_fd);

    max_fd = 0;
}

int TcpServer::Poll(fd_set& in_rset, fd_set& in_wset) {
    int ret;

    if (!sv_valid)
        return -1;

    // Look for new connections we need to accept
    int accept_fd = 0;
    if (FD_ISSET(serv_fd, &in_rset)) {
        // Accept an inbound connection.  This is non-fatal if it fails
        if ((accept_fd = Accept()) < 0)
            return 0;
        // Validate them and see if they're allowed to remain
        // Bounce back a 0 so we can log the refusal
        if (Validate(accept_fd) < 0) {
            KillConnection(accept_fd);
            return 0;
        }
    }

    // Handle input and output, dispatching to our other functions so we can
    // be overridden
    for (unsigned int x = 0; x <= max_fd; x++) {
        // Handle reading data.  Accept() should have made them a 
        // ringbuffer.
        if (FD_ISSET(x, &in_rset) && FD_ISSET(x, &server_fdset)) {
            if ((ret = ReadBytes(x)) < 0) {
                KillConnection(x);
                continue;
            }

            // Try to parse it.  We only do this when its changed since
            // if it couldn't parse a fragment before it's not going to be
            // able to parse a fragment still
            if (ret > 0 && ParseData(x) < 0) {
                KillConnection(x);
                continue;
            }
        }

        // Handle writing data.  The write FD would never have gotten set
        // for checking if there wasn't data in the ring buffer, so we don't
        // have to check for that.
        if (FD_ISSET(x, &in_wset) && FD_ISSET(x, &server_fdset)) {
            if ((ret = WriteBytes(x)) < 0)
                continue;
        }

    }

    return 1;
}

int TcpServer::TcpAccept() {
    unsigned int new_fd;
    struct sockaddr_in client_addr;
#ifdef HAVE_SOCKLEN_T
    socklen_t client_len;
#else
    int client_len;
#endif

    memset(&client_addr, 0, sizeof(struct sockaddr_in));
    client_len = sizeof(struct sockaddr_in);

    // Socket accept
    if ((new_fd = accept(serv_fd, (struct sockaddr *) &client_addr,
                         &client_len)) < 0) {
        snprintf(errstr, 1024, "TCP server accept() failed: %s",
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Bail right now if we have too many connections
    if (FetchNumClients() >= (int) max_clients) {
        snprintf(errstr, 1024, "TCP server maximum clients (%d) already reached.",
                 max_clients);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        close(new_fd);
        return -1;
    }

    // Set it to nonblocking
    int save_mode = fcntl(new_fd, F_GETFL, 0);
    fcntl(new_fd, F_SETFL, save_mode | O_NONBLOCK);
    
    if (new_fd > max_fd)
        max_fd = new_fd;

    FD_SET(new_fd, &server_fdset);

    // There should never be overlapping fds and there should never be
    // remnants of an old person here, so we'll make the connection 
    // lightweight and not do more tree searching.  If this ever proves
    // wrong, we need to reevaluate
    write_buf_map[new_fd] = new RingBuffer(SRV_RING_LEN);
    read_buf_map[new_fd] = new RingBuffer(SRV_RING_LEN);

    return new_fd;
}

int TcpServer::Accept() {
    // Just handle the TCP acceptance stuff
    return TcpAccept();
}

int TcpServer::ValidateIPFilter(int in_fd) {
    struct sockaddr_in client_addr;
#ifdef HAVE_SOCKLEN_T
    socklen_t client_len;
#else
    int client_len;
#endif

    memset(&client_addr, 0, sizeof(struct sockaddr_in));
    client_len = sizeof(struct sockaddr_in);

    if (getsockname(in_fd, (struct sockaddr *) &client_addr, &client_len) < 0) {
        snprintf(errstr, 1024, "TCP server unable to get sockname for validation: %s",
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // No filtering = valid
    if (ipfilter_vec.size() == 0)
        return 1;

    char inhost[16];
    snprintf(inhost, 16, "%s", inet_ntoa(client_addr.sin_addr));

    int legal_ip = 0;
    for (unsigned int ibvi = 0; ibvi < ipfilter_vec.size(); ibvi++) {
        if ((client_addr.sin_addr.s_addr & ipfilter_vec[ibvi]->mask.s_addr) ==
            ipfilter_vec[ibvi]->network.s_addr) {
            legal_ip = 1;
            break;
        }
    }

    if (legal_ip == 0) {
        snprintf(errstr, 1024, "TCP server client fd %d from untrusted host %s refused",
                 in_fd, inhost);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        KillConnection(in_fd);
        return -1;
    }

    return 1;
}

int TcpServer::Validate(int in_fd) {
    // Just validate based on IP block
    return ValidateIPFilter(in_fd);
}

int TcpServer::ReadBytes(int in_fd) {
    uint8_t recv_bytes[1024];
    int ret;

    if ((ret = read(in_fd, recv_bytes, 1024)) < 0) {
        snprintf(errstr, 1024, "TCP server client fd %d read() error: %s", 
                 in_fd, strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (ret == 0) {
        snprintf(errstr, 1024, "TCP server client fd %d read() returned end of file",
                 in_fd);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (read_buf_map[in_fd]->InsertData(recv_bytes, ret) == 0) {
        snprintf(errstr, 1024, "TCP server client fd %d read error, ring buffer full",
                 in_fd);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    return ret;
}

int TcpServer::ParseData(int in_fd) {
    snprintf(errstr, 1024, "TCP server: No data parser.  How did this get called?");
    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
    return -1;
}

int TcpServer::WriteBytes(int in_fd) {
    uint8_t dptr[1024];
    int dlen, ret;

    // This can't get called on invalid fds, so save some time and
    // don't check
    write_buf_map[in_fd]->FetchPtr(dptr, 1024, &dlen);

    if ((ret = write(in_fd, dptr, dlen)) <= 0) {
        snprintf(errstr, 1024, "TCP server: Killing client fd %d write error %s",
                 in_fd, strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        KillConnection(in_fd);
        return -1;
    }

    write_buf_map[in_fd]->MarkRead(ret);

    return ret;
}

