/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include "tcpsslapclient.h"
#include "endian_magic.h"
#include "ifcontrol.h"
#include "iwcontrol.h"

void TcpSSLApMessageClient::ProcessMessage(string in_msg, int in_flags) {
    if (in_flags & MSGFLAG_LOCAL)
        return;

    int proto_flags = 0;
    if (in_flags & MSGFLAG_DEBUG)
        proto_flags |= AP_MSGFLAG_DEBUG;
    if (in_flags & MSGFLAG_INFO)
        proto_flags |= AP_MSGFLAG_INFO;
    if (in_flags & MSGFLAG_ERROR)
        proto_flags |= AP_MSGFLAG_ERROR;
    if (in_flags & MSGFLAG_FATAL)
        proto_flags |= AP_MSGFLAG_FATAL;

    // Propogate it to the server
    cli->SendMessagePacket(in_msg.c_str(), proto_flags);
}

int SSLApClientUpdateStatsTimer(Timetracker::timer_event *evt, void *parm) {
    // Just kick off the function inside the server and don't repeat
    TcpSSLApClient *cli = (TcpSSLApClient *) parm;

    if (cli->UpdateStats() < 0)
        return 0;

    return 1;
}

int SSLApClientConnectKismetTimer(Timetracker::timer_event *evt, void *parm) {
    TcpSSLApClientKismetTimerRec *tr = (TcpSSLApClientKismetTimerRec *) parm;

    // Bounce into the object
    tr->cli->ConnectKismetInterface(tr->ifnum);

    delete tr;

    return 0;
}

int SSLApClientSendHeloTimer(Timetracker::timer_event *evt, void *parm) {
    TcpSSLApClient *cli = (TcpSSLApClient *) parm;

    if (cli->SendHeloPacket() < 0)
        return 0;

    return 1;
}

int SSLApClientInterfaceThrashTimer(Timetracker::timer_event *evt, void *parm) {
    TcpSSLApClient *cli = (TcpSSLApClient *) parm;

    cli->ThrashTick();
    
    return 1;
}

TcpSSLApClient::TcpSSLApClient() {
    tx_seqid = 0;
    rx_seqid = 0;
    station_id = 254;
    station_name[0] = '\0';
    state = PS_PENDING_REGISTRATION;
    stats_event_id = 0;
    helo_event_id = 0;
    thrash_event_id = 0;
    kisport = 2501;
    ssid[0] = '\0';

}

int TcpSSLApClient::Connect(const char *in_remotehost, short int in_port) {
    int ret;

    // Register the thrash timer
    if (thrash_event_id == 0)
        thrash_event_id = timetracker->RegisterTimer(SERVER_TIMESLICES_SEC,
                                                     NULL, 1, &SSLApClientInterfaceThrashTimer, 
                                                     (void *) this);

    if ((ret = TcpSSLClient::Connect(in_remotehost, in_port)) < 0)
        return ret;

    // They'd better have configured us before calling this...
    return SendAnnouncePacket();
}

unsigned int TcpSSLApClient::MergeSet(fd_set in_rset, fd_set in_wset, 
                                      unsigned int in_max_fd,
                                      fd_set *out_rset, fd_set *out_wset) {
    unsigned int newmax = in_max_fd;

    // Merge with the clients
    for (unsigned int x = 0; x < 4; x++) {
        if (interfaces[x].kisclient != NULL &&
            interfaces[x].kisclient->Valid()) {
            newmax = interfaces[x].kisclient->MergeSet(in_rset, in_wset, newmax,
                                                       out_rset, out_wset);
        }
    }

    // Call the parent
    return TcpSSLClient::MergeSet(*out_rset, *out_wset, newmax, out_rset, out_wset);
}


int TcpSSLApClient::Poll(fd_set& in_rset, fd_set& in_wset) {
    int ret = 0;

    // Poll the kismet clients
    for (unsigned int x = 0; x < 4; x++) {
        if (interfaces[x].kisclient != NULL &&
            interfaces[x].kisclient->Valid()) {
            if ((ret = interfaces[x].kisclient->Poll(in_rset, in_wset)) < 0) {
                snprintf(errstr, 1024, "Interface '%s' Kismet failed",
                         interfaces[x].ifname);
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);

                // Flag it as a broken interface
                interfaces[x].cf_bitmap |= AP_IFFLAG_ERROR;

                // Kick off a kismet restart here to try to fix it
                interfaces[x].kisclient->KillConnection();

                interfaces[x].kismet_pending = 0;
                interfaces[x].dirty = 1;

                if (interfaces[x].kismet_pid != 0)
                    kill(interfaces[x].kismet_pid, SIGKILL);
                interfaces[x].kismet_pid = 0;

                ActivateInterfaces();

            }

            if (ret -= 0) {
                if (time(0) - interfaces[x].kismet_time > 10) {
                    snprintf(errstr, 1024, "Interface '%s' didn't hear from Kismet "
                             "in more than 10 seconds, something is wrong.",
                             interfaces[x].ifname);
                    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                    // Flag it as a broken interface
                    interfaces[x].cf_bitmap |= AP_IFFLAG_ERROR;
                   
                    // Kick off a kismet restart here to try to fix it
                    interfaces[x].kisclient->KillConnection();

                    interfaces[x].kismet_pending = 0;
                    interfaces[x].dirty = 1;

                    if (interfaces[x].kismet_pid != 0)
                        kill(interfaces[x].kismet_pid, SIGKILL);
                    interfaces[x].kismet_pid = 0;

                    ActivateInterfaces();
                }
            } else {
                interfaces[x].kismet_time = time(0);

                if (interfaces[x].kisclient->FetchNetworkDirty()) {
                    vector<TcpKismetClient::network_rec *> nvec;
                    interfaces[x].kisclient->FetchNetworks(&nvec);

                    // Send new and updated networks to the server to handle
                    for (unsigned int nx = 0; nx < nvec.size(); nx++) {
                        if (nvec[nx]->type != TcpKismetClient::network_ap)
                            continue;

                        if (nvec[nx]->last_time >= interfaces[x].last_kis_network)
                            SendKisNetworkPacket(nvec[nx]);
                    }

                    interfaces[x].last_kis_network = time(0);
                }

                if (interfaces[x].kisclient->FetchAlertDirty()) {
                    vector<TcpKismetClient::alert_rec *> avec;
                    interfaces[x].kisclient->FetchAlerts(&avec);

                    // Send new alerts to the server to handle
                    for (int ax = (int) (avec.size() - 1); ax >= 0; --ax) {
                        if (avec[ax]->timestamp < interfaces[x].last_kis_alert)
                            break;

                        SendKisAlertPacket(avec[ax]);
                    }

                    interfaces[x].last_kis_alert = interfaces[x].kismet_time;
                }
            }
        }
    }

    return TcpSSLClient::Poll(in_rset, in_wset);
}

int TcpSSLApClient::RegisterEnvironment(const char *in_name,
                                        const char *in_if0,
                                        const char *in_if1,
                                        const char *in_if2,
                                        const char *in_if3,
                                        unsigned int in_zone,
                                        const char *in_kispath,
                                        const char *in_smitepath,
                                        const char *in_aphelperpath) {

    strncpy(station_name, in_name, 32);
    zone = in_zone;

    // Fetch the info about all the interfaces
    int navail = 0;
    if (strlen(in_if0) > 0) {
        if (Ifconfig_Get_Hwaddr(in_if0, errstr, interfaces[0].innate_hwaddr) < 0) {
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        } else {
            navail++;
            strncpy(interfaces[0].ifname, in_if0, 32);
            interfaces[0].cf_bitmap |= (AP_IFFLAG_AVAILABLE);
        }
    }
    if (strlen(in_if1) > 0) {
        if (Ifconfig_Get_Hwaddr(in_if1, errstr, interfaces[1].innate_hwaddr) < 0) {
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        } else {
            navail++;
            strncpy(interfaces[1].ifname, in_if1, 32);
            interfaces[1].cf_bitmap |= (AP_IFFLAG_AVAILABLE);
        }
    }
    if (strlen(in_if2) > 0) {
        if (Ifconfig_Get_Hwaddr(in_if2, errstr, interfaces[2].innate_hwaddr) < 0) {
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        } else {
            navail++;
            strncpy(interfaces[2].ifname, in_if2, 32);
            interfaces[2].cf_bitmap |= (AP_IFFLAG_AVAILABLE);
        }
    }
    if (strlen(in_if3) > 0) {
        if (Ifconfig_Get_Hwaddr(in_if3, errstr, interfaces[3].innate_hwaddr) < 0) {
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        } else {
            navail++;
            strncpy(interfaces[3].ifname, in_if3, 32);
            interfaces[3].cf_bitmap |= (AP_IFFLAG_AVAILABLE);
        }
    }

    if (navail == 0) {
        snprintf(errstr, 1024, "No available interfaces specified, no reason to "
                 "continue.");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (navail == 0) {
        msgbus->InjectMessage("No requested interfaces available.", 
                              (MSGFLAG_ERROR | MSGFLAG_FATAL));
        return -1;
    }

    kispath = strdup(in_kispath);
    smitepath = strdup(in_smitepath);
    aphelperpath = strdup(in_aphelperpath);

    state = PS_PENDING_ANNOUNCE;

    // Start updating stats now that we can
    if (stats_event_id == 0)
        stats_event_id = timetracker->RegisterTimer(SERVER_TIMESLICES_SEC / 2,
                                                    NULL, 1, 
                                                    &SSLApClientUpdateStatsTimer, 
                                                    (void *) this);
    
    return 1;
}

typedef struct {
    unsigned long long int rx_bytecount;
    unsigned long long int tx_bytecount;
    unsigned long long int rx_packetcount;
    unsigned long long int tx_packetcount;
} _procnetdevdat;

int TcpSSLApClient::UpdateStats() {
    HostapStatFile sf;
    char statname[512];
    int ret;
    map<string, _procnetdevdat> pnd_map;

    // Process /proc/net/dev and put it in a local map
    FILE *pnd;

    if ((pnd = fopen("/proc/net/dev", "r")) == NULL) {
        snprintf(errstr, 1024, "TcpSSLApClient unable to open /proc/net/dev to read statistics: %s",
                 strerror(errno));
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
    } else {
        unsigned long int junk;

        // scrap the first 2 lines of file headers
        fgets(errstr, 1024, pnd);
        fgets(errstr, 1024, pnd);

        do {
            char tifname[32];
            _procnetdevdat dat;
            if (fscanf(pnd, "%32[^:]: %llu %llu %lu %lu %lu %lu %lu %lu %llu %llu %lu %lu %lu %lu %lu %lu\n",
                       tifname, &dat.rx_bytecount, &dat.rx_packetcount, 
                       &junk, &junk, &junk, &junk, &junk, &junk,
                       &dat.tx_bytecount, &dat.tx_packetcount,
                       &junk, &junk, &junk, &junk, &junk, &junk) < 17) {
                snprintf(errstr, 1024, "TcpSSLApClient unable to parse line in /proc/net/dev, skipping.");
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            }

            pnd_map[StrStrip(tifname)] = dat;
            //printf("debug - '%s' - %llu %llu %llu %llu\n", ifname, rx_bytes, tx_bytes, rx_packets, tx_packets);
        } while (!feof(pnd));
        fclose(pnd);
    }

    for (unsigned int x = 0; x < 4; x++) {
        // Reap completed rawtx processes, look for overdue ones and try to
        // kill them, update config bitmaps
        if (interfaces[x].rawtx_pid > 0) {
            if ((ret = wait4(interfaces[x].rawtx_pid, NULL, WNOHANG, NULL)) ==
                interfaces[x].rawtx_pid) {
                printf("debug - reaped rawtx process\n");
                interfaces[x].rawtx_pid = 0;
                interfaces[x].cf_bitmap &= ~AP_IFFLAG_BUSY;
            } else if (ret < 0) {
                snprintf(errstr, 1024, "TcpSSLApClient wait4() rawtx "
                         "interface %s pid %d failed: %s", interfaces[x].ifname,
                         interfaces[x].rawtx_pid, strerror(errno));
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            } else if ((interfaces[x].rawtx_time + 120) > time(0)) {
                snprintf(errstr, 1024, "TcpSSLApClient rawtx interface %s "
                         "pid %d failed to complete within 120 seconds, sending "
                         "SIGKILL", interfaces[x].ifname, interfaces[x].rawtx_pid);
                msgbus->InjectMessage(errstr, MSGFLAG_INFO);
                kill(interfaces[x].rawtx_pid, SIGKILL);
            }
        }
        
        // Skip stuff that doesn't get updated interface statistics
        if ((interfaces[x].cf_bitmap & AP_ACTIVEAP) != AP_ACTIVEAP) {
            continue;
        }

        snprintf(statname, 512, "/proc/net/hostap/%s/stats", 
                 interfaces[x].ifname);

        if (sf.ParseStats(statname) < 0) {
            printf("debug - inject error here\n");
            // mesbus.InjectMessage(sf.FetchError(), MSGFLAG_ERROR);
            // return -1;
        }

        // Add up all the relevant packet counts and error counts
        interfaces[x].framecount = sf.FetchStat("TxUnicastFrames") + 
            sf.FetchStat("TxMulticastframes") +
            sf.FetchStat("RxUnicastFrames") + 
            sf.FetchStat("RxMulticastframes");
        interfaces[x].errorcount = sf.FetchStat("RxFCSErrors");

        // Find all the stations associated and count them
        snprintf(statname, 512, "/proc/net/hostap/%s", 
                 interfaces[x].ifname);
        if (HostAPFetchStations(statname, errstr, &(interfaces[x].station_map)) < 0) {
            printf("debug - inject error here\n");
            // mesbus.InjectMessage(errstr, MSGFLAG_ERROR);
            // return -1;
        }

        interfaces[x].clientcount = interfaces[x].station_map.size();

        map<string, _procnetdevdat>::iterator pndi = pnd_map.find(interfaces[x].ifname);
        if (pndi != pnd_map.end()) {
            interfaces[x].rx_packetcount = pndi->second.rx_packetcount;
            interfaces[x].rx_bytecount = pndi->second.rx_bytecount;
            interfaces[x].tx_packetcount = pndi->second.tx_packetcount;
            interfaces[x].tx_bytecount = pndi->second.tx_bytecount;
        }

    }

    if (FetchSysLoadAvg(&load_maj, &load_min) < 0) {
        snprintf(errstr, 1024, "Fetching load average from /proc/loadavg failed.");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }
    
    return 1;
}

int TcpSSLApClient::SendAnnouncePacket() {
    if (!cl_valid || state == PS_PENDING_REGISTRATION)
        return 0;

    stream_frame frameheader;
    ap_announce_packet annpayload;

    frameheader.frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    frameheader.frame_type = FRAMETYPE_ANNOUNCE;
    frameheader.frame_len = (uint32_t) htonl(FRAMELEN_ANNOUNCE);
    frameheader.payload = &annpayload;

    strncpy((char *) annpayload.name, station_name, 32);

    // Send the innate address, not whatever we got set to
    memcpy(annpayload.bssid0, interfaces[0].innate_hwaddr, 6);
    memcpy(annpayload.bssid1, interfaces[1].innate_hwaddr, 6);
    memcpy(annpayload.bssid2, interfaces[2].innate_hwaddr, 6);
    memcpy(annpayload.bssid3, interfaces[3].innate_hwaddr, 6);

    annpayload.cf_bitmap0 = (uint32_t) htonl(interfaces[0].cf_bitmap);
    annpayload.cf_bitmap1 = (uint32_t) htonl(interfaces[1].cf_bitmap);
    annpayload.cf_bitmap2 = (uint32_t) htonl(interfaces[2].cf_bitmap);
    annpayload.cf_bitmap3 = (uint32_t) htonl(interfaces[3].cf_bitmap);

    snprintf((char *) annpayload.ifname0, 32, interfaces[0].ifname);
    snprintf((char *) annpayload.ifname1, 32, interfaces[1].ifname);
    snprintf((char *) annpayload.ifname2, 32, interfaces[2].ifname);
    snprintf((char *) annpayload.ifname3, 32, interfaces[3].ifname);

    annpayload.zone = zone;

    if (QueuePacket(&frameheader) < 0)
        return -1;

    state = PS_PENDING_CONFIGURE;

    snprintf(errstr, 1024, "Sent announce frame to AP coordinator.");
    msgbus->InjectMessage(errstr, MSGFLAG_INFO | MSGFLAG_LOCAL);
    
    return 1;
}

int TcpSSLApClient::SendHeloPacket() {
    if (!cl_valid || state != PS_RUNNING)
        return 0;

    stream_frame frameheader;
    ap_helo_packet helopayload;
    struct timeval tm;

    // Fill in the stuff
    frameheader.frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    frameheader.frame_type = FRAMETYPE_HELO;
    frameheader.frame_len = (uint32_t) htonl(FRAMELEN_HELO);
    frameheader.payload = &helopayload;

    gettimeofday(&tm, NULL);
    helopayload.station_timestamp_sec = (uint32_t) htonl(tm.tv_sec);
    helopayload.station_timestamp_usec = (uint32_t) htonl(tm.tv_usec);
    
    helopayload.load_avg_major = load_maj;
    helopayload.load_avg_minor = load_min;

    helopayload.clientcount0 = (uint32_t) htonl(interfaces[0].clientcount);
    helopayload.clientcount1 = (uint32_t) htonl(interfaces[1].clientcount);
    helopayload.clientcount2 = (uint32_t) htonl(interfaces[2].clientcount);
    helopayload.clientcount3 = (uint32_t) htonl(interfaces[3].clientcount);

    helopayload.framecount0 = (uint64_t) kis_hton64(interfaces[0].framecount);
    helopayload.framecount1 = (uint64_t) kis_hton64(interfaces[1].framecount);
    helopayload.framecount2 = (uint64_t) kis_hton64(interfaces[2].framecount);
    helopayload.framecount3 = (uint64_t) kis_hton64(interfaces[3].framecount);

    helopayload.errorcount0 = (uint64_t) kis_hton64(interfaces[0].errorcount);
    helopayload.errorcount1 = (uint64_t) kis_hton64(interfaces[1].errorcount);
    helopayload.errorcount2 = (uint64_t) kis_hton64(interfaces[2].errorcount);
    helopayload.errorcount3 = (uint64_t) kis_hton64(interfaces[3].errorcount);

    helopayload.txpacketcount0 = (uint64_t) kis_hton64(interfaces[0].tx_packetcount);
    helopayload.txpacketcount1 = (uint64_t) kis_hton64(interfaces[1].tx_packetcount);
    helopayload.txpacketcount2 = (uint64_t) kis_hton64(interfaces[2].tx_packetcount);
    helopayload.txpacketcount3 = (uint64_t) kis_hton64(interfaces[3].tx_packetcount);

    helopayload.rxpacketcount0 = (uint64_t) kis_hton64(interfaces[0].rx_packetcount);
    helopayload.rxpacketcount1 = (uint64_t) kis_hton64(interfaces[1].rx_packetcount);
    helopayload.rxpacketcount2 = (uint64_t) kis_hton64(interfaces[2].rx_packetcount);
    helopayload.rxpacketcount3 = (uint64_t) kis_hton64(interfaces[3].rx_packetcount);
    
    helopayload.txbytecount0 = (uint64_t) kis_hton64(interfaces[0].tx_bytecount);
    helopayload.txbytecount1 = (uint64_t) kis_hton64(interfaces[1].tx_bytecount);
    helopayload.txbytecount2 = (uint64_t) kis_hton64(interfaces[2].tx_bytecount);
    helopayload.txbytecount3 = (uint64_t) kis_hton64(interfaces[3].tx_bytecount);

    helopayload.rxbytecount0 = (uint64_t) kis_hton64(interfaces[0].rx_bytecount);
    helopayload.rxbytecount1 = (uint64_t) kis_hton64(interfaces[1].rx_bytecount);
    helopayload.rxbytecount2 = (uint64_t) kis_hton64(interfaces[2].rx_bytecount);
    helopayload.rxbytecount3 = (uint64_t) kis_hton64(interfaces[3].rx_bytecount);

    helopayload.cf_bitmap0 = (uint32_t) htonl(interfaces[0].cf_bitmap);
    helopayload.cf_bitmap1 = (uint32_t) htonl(interfaces[1].cf_bitmap);
    helopayload.cf_bitmap2 = (uint32_t) htonl(interfaces[2].cf_bitmap);
    helopayload.cf_bitmap3 = (uint32_t) htonl(interfaces[3].cf_bitmap);

    // Queue it.  We copy it into our ringbuffers so this is safe with local vars
    if (QueuePacket(&frameheader) < 0)
        return -1;

    return 1;
}

int TcpSSLApClient::SendMessagePacket(const char *in_msg, int in_flags) {
    if (!cl_valid)
        return 0;

    stream_frame frameheader;
    uint8_t *rawpayload;

    frameheader.frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    frameheader.frame_type = FRAMETYPE_MESSAGE;
    frameheader.frame_len = (uint32_t) htonl(FRAMELEN_MESSAGE_P + strlen(in_msg) + 1);

    rawpayload = new uint8_t[FRAMELEN_MESSAGE_P + strlen(in_msg) + 1];

    frameheader.payload = rawpayload;

    // Flags and size
    ((ap_message_packet *) rawpayload)->msgflags = htonl(in_flags);
    ((ap_message_packet *) rawpayload)->msglen = htonl(strlen(in_msg) + 1);

    // This is just stupid
    strncpy((char *) &(((ap_message_packet *) rawpayload)->message), in_msg,
            strlen(in_msg) + 1);

    if (QueuePacket(&frameheader) < 0)
        return -1;

    delete[] rawpayload;
    
    return 1;
}

int TcpSSLApClient::SendKisAlertPacket(TcpKismetClient::alert_rec *in_alert) {
    if (!cl_valid)
        return 0;

    stream_frame frameheader;
    ap_kisalert_packet apayload;

    // Fill in the stuff
    frameheader.frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    frameheader.frame_type = FRAMETYPE_KISALERT;
    frameheader.frame_len = (uint32_t) htonl(FRAMELEN_KISALERT);
    frameheader.payload = &apayload;

    apayload.type = (uint8_t) in_alert->type;
    
    for (unsigned int x = 0; x < 6; x++) {
        apayload.bssid_mac[x] = in_alert->bssid_mac[x];
        apayload.source_mac[x] = in_alert->source_mac[x];
        apayload.dest_mac[x] = in_alert->dest_mac[x];
        apayload.other_mac[x] = in_alert->other_mac[x];
    }

    apayload.channel = (uint16_t) htons(in_alert->channel);
    apayload.timestamp = (uint32_t) htonl(in_alert->timestamp);

    if (QueuePacket(&frameheader) < 0)
        return -1;

    return 1;
}

int TcpSSLApClient::SendKisNetworkPacket(TcpKismetClient::network_rec *in_net) {
    if (!cl_valid)
        return 0;

    stream_frame frameheader;
    ap_kisnetwork_packet npayload;

    // Fill in the stuff
    frameheader.frame_sentinel = (uint32_t) htonl(STREAM_SENTINEL);
    frameheader.frame_type = FRAMETYPE_KISNETWORK;
    frameheader.frame_len = (uint32_t) htonl(FRAMELEN_KISNETWORK);
    frameheader.payload = &npayload;

    npayload.type = (uint8_t) in_net->type;
    
    for (unsigned int x = 0; x < 6; x++) {
        npayload.bssid_mac[x] = in_net->bssid_mac[x];
    }

    strncpy((char *) npayload.ssid, in_net->ssid.c_str(), 32);

    npayload.attributes = (uint32_t) htonl(in_net->attributes);

    npayload.channel = (uint16_t) htons(in_net->channel);

    npayload.first_time = (uint32_t) htonl(in_net->first_time);
    npayload.last_time = (uint32_t) htonl(in_net->last_time);

    if (QueuePacket(&frameheader) < 0)
        return -1;

    return 1;

}

int TcpSSLApClient::ParseData() {
    int ret = 0;

    // Look for something long enough to hold a frame
    while (read_buf->FetchLen() >= (int) FRAMELEN_HEADER) {
        // Map it into the frame header
        stream_frame fhead;
        int flen;
    
        read_buf->FetchPtr((uint8_t *) &fhead, FRAMELEN_HEADER, &flen);

        // Validate the frame header
        if (fhead.frame_sentinel != htonl(STREAM_SENTINEL)) {
            snprintf(errstr, 1024, "TCP SSL AP client got an invalid frame sentinel, "
                     "closing connection");
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            return -1;
        }

        fhead.frame_len = (uint32_t) ntohl(fhead.frame_len);

        // Do we have enough for the payload?
        if (read_buf->FetchLen() >= (int) (FRAMELEN_HEADER + fhead.frame_len)) {
            // Allocate a payload frame
            uint8_t *payload = new uint8_t[fhead.frame_len];

            // Mark the header as read in the ring
            read_buf->MarkRead(FRAMELEN_HEADER);

            // Fetch the payload and mark it read
            read_buf->FetchPtr(payload, fhead.frame_len, &flen);
            read_buf->MarkRead(fhead.frame_len);

            // Handle packet types
            if (fhead.frame_type == FRAMETYPE_CONFIGURE) {
                ret = ParseConfigure((mn_configure_packet *) payload);
            } else if (fhead.frame_type == FRAMETYPE_ACTIVATE) {
                ret = ParseActivate((mn_activate_packet *) payload);
            } else if (fhead.frame_type == FRAMETYPE_CHBSSID) {
                ret = ParseBSSID((mn_bssid_packet *) payload);
            } else if (fhead.frame_type == FRAMETYPE_SMITE) {
                ret = ParseSmite((mn_smite_packet *) payload);
            } else {
                snprintf(errstr, 1024, "TCP SSL AP client got unknown frame type %d, "
                         "closing connection", fhead.frame_type);
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                ret = -1; 
            }

            delete[] payload;
        } else {
            // Drop out of the loop
            break;
        }

    }

    return ret;
}

int TcpSSLApClient::ParseConfigure(mn_configure_packet *in_conf) {
    if (state != PS_PENDING_CONFIGURE && state != PS_RUNNING) {
        snprintf(errstr, 1024, "TCPSSLAP client got out-of-order AP configure "
                 "frame, something is wrong, closing connection.");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    rx_seqid = kis_ntoh64(in_conf->sequence_id) + 1;

    station_id = (unsigned int) ntohs(in_conf->station_id);

    strncpy(ssid, (char *) in_conf->ssid, 32);
    
    // Back up info
    interfaces[0].old_channel = interfaces[0].channel;
    interfaces[1].old_channel = interfaces[1].channel;
    interfaces[2].old_channel = interfaces[2].channel;
    interfaces[3].old_channel = interfaces[3].channel;
    interfaces[0].old_cf_bitmap = interfaces[0].cf_bitmap;
    interfaces[1].old_cf_bitmap = interfaces[1].cf_bitmap;
    interfaces[2].old_cf_bitmap = interfaces[2].cf_bitmap;
    interfaces[3].old_cf_bitmap = interfaces[3].cf_bitmap;

    interfaces[0].channel = ntohs(in_conf->ifchannel0);
    interfaces[1].channel = ntohs(in_conf->ifchannel1);
    interfaces[2].channel = ntohs(in_conf->ifchannel2);
    interfaces[3].channel = ntohs(in_conf->ifchannel3);

    interfaces[0].cf_bitmap = ntohl(in_conf->cf_bitmap0);
    interfaces[1].cf_bitmap = ntohl(in_conf->cf_bitmap1);
    interfaces[2].cf_bitmap = ntohl(in_conf->cf_bitmap2);
    interfaces[3].cf_bitmap = ntohl(in_conf->cf_bitmap3);

    // Never let anyone tell you you're bad or too busy
    interfaces[0].cf_bitmap &= ~(AP_IFFLAG_ERROR | AP_IFFLAG_BUSY);
    interfaces[1].cf_bitmap &= ~(AP_IFFLAG_ERROR | AP_IFFLAG_BUSY);
    interfaces[2].cf_bitmap &= ~(AP_IFFLAG_ERROR | AP_IFFLAG_BUSY);
    interfaces[3].cf_bitmap &= ~(AP_IFFLAG_ERROR | AP_IFFLAG_BUSY);

    if (memcmp(interfaces[0].hwaddr, in_conf->bssid0, 6)) {
        memcpy(interfaces[0].hwaddr, in_conf->bssid0, 6);
        interfaces[0].macdirty = 0;
    }
    if (memcmp(interfaces[1].hwaddr, in_conf->bssid1, 6)) {
        memcpy(interfaces[1].hwaddr, in_conf->bssid1, 6);
        interfaces[1].macdirty = 0;
    }
    if (memcmp(interfaces[2].hwaddr, in_conf->bssid2, 6)) {
        memcpy(interfaces[2].hwaddr, in_conf->bssid2, 6);
        interfaces[2].macdirty = 0;
    }
    if (memcmp(interfaces[3].hwaddr, in_conf->bssid3, 6)) {
        memcpy(interfaces[3].hwaddr, in_conf->bssid3, 6);
        interfaces[3].macdirty = 0;
    }
        
    interfaces[0].dirty = 1;
    interfaces[1].dirty = 1;
    interfaces[2].dirty = 1;
    interfaces[3].dirty = 1;

    interfaces[0].kismet_pid = 0;
    interfaces[1].kismet_pid = 0;
    interfaces[2].kismet_pid = 0;
    interfaces[3].kismet_pid = 0;

    state = PS_CONFIGURED;
    
    snprintf(errstr, 1024, "AP coordinator has configured us, waiting for activation.");
    msgbus->InjectMessage(errstr, MSGFLAG_INFO | MSGFLAG_LOCAL);

    return 1;
}

int TcpSSLApClient::ParseActivate(mn_activate_packet *in_act) {
    snprintf(errstr, 1024, "AP coordinator has activated us.");
    msgbus->InjectMessage(errstr, MSGFLAG_INFO | MSGFLAG_LOCAL);

    if (state != PS_CONFIGURED) {
        snprintf(errstr, 1024, "TCP SSL AP client got out-of-order AP activate "
                 "frame, something is wrong, closing connection.");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    unsigned long int in_seqid = (long int) kis_ntoh64(in_act->sequence_id);
    if (in_seqid < rx_seqid ||
        in_seqid > rx_seqid + STREAM_SEQ_WINDOW) {
        snprintf(errstr, 1024, "Data recieved from server out-of-sequence, got "
                 "%lu expected %lu", in_seqid, rx_seqid);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    rx_seqid = in_seqid + 1;

    state = PS_RUNNING;

    if (helo_event_id == 0)
        helo_event_id = timetracker->RegisterTimer(SERVER_TIMESLICES_SEC,
                                                   NULL, 1, 
                                                   &SSLApClientSendHeloTimer, 
                                                   (void *) this);
    return ActivateInterfaces();
}

int TcpSSLApClient::ParseBSSID(mn_bssid_packet *in_act) {
    unsigned long int in_seqid = (long int) kis_ntoh64(in_act->sequence_id);
    if (in_seqid < rx_seqid ||
        in_seqid > rx_seqid + STREAM_SEQ_WINDOW) {
        snprintf(errstr, 1024, "Data recieved from server out-of-sequence, got "
                 "%lu expected %lu", in_seqid, rx_seqid);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    rx_seqid = in_seqid + 1;

    // See if we're an AP
    if (in_act->if_index > 4) {
        snprintf(errstr, 1024, "BSSID Change: Illegal interface index %u", 
                 in_act->if_index);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Try to set the bssid
    if ((interfaces[in_act->if_index].cf_bitmap & AP_ACTIVEAP) == AP_ACTIVEAP) {
        if (Iwconfig_Set_IntPriv(interfaces[in_act->if_index].ifname, "maccmd", 
                                 4, 0, errstr) < 0) {
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        }
        if (Ifconfig_Set_Hwaddr(interfaces[in_act->if_index].ifname, errstr,
                                interfaces[in_act->if_index].hwaddr) < 0) {
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            interfaces[in_act->if_index].cf_bitmap |= AP_IFFLAG_ERROR;
            interfaces[in_act->if_index].macdirty = 1;
        }
    }

    return 1;
}

int TcpSSLApClient::ParseSmite(mn_smite_packet *in_act) {
    unsigned long int in_seqid = (long int) kis_ntoh64(in_act->sequence_id);
    if (in_seqid < rx_seqid ||
        in_seqid > rx_seqid + STREAM_SEQ_WINDOW) {
        snprintf(errstr, 1024, "Data recieved from server out-of-sequence, got "
                 "%lu expected %lu", in_seqid, rx_seqid);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    rx_seqid = in_seqid + 1;

    // See if we're an AP
    if (in_act->if_index > 4) {
        snprintf(errstr, 1024, "Smite: Illegal interface index %u", 
                 in_act->if_index);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if ((interfaces[in_act->if_index].cf_bitmap & AP_IFFLAG_RAWTX) == 0) {
        snprintf(errstr, 1024, "Smite: Illegal interface index %u, "
                 "not configured as rawtx", in_act->if_index);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if ((interfaces[in_act->if_index].cf_bitmap & AP_IFFLAG_BUSY) != 0) {
        snprintf(errstr, 1024, "Smite: Illegal interface index %u, "
                 "interface flagged as busy", in_act->if_index);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (interfaces[in_act->if_index].rawtx_pid > 0) {
        snprintf(errstr, 1024, "Smite: Illegal interface index %u, "
                 "nonzero rawtx pid - we're still busy running the last one", 
                 in_act->if_index);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    int channel = ntohs(in_act->channel);

    // Spawn the rawtx agent
    char cmdline[1024];
    snprintf(cmdline, 1024, "%s -i %s -b %02X:%02X:%02X:%02X:%02X:%02X -c %d -S %s",
             smitepath, interfaces[in_act->if_index].ifname,
             in_act->bssid[0], in_act->bssid[1], in_act->bssid[2],
             in_act->bssid[3], in_act->bssid[4], in_act->bssid[5],
             channel, ssid);
    interfaces[in_act->if_index].rawtx_pid = ExecSysCmd(cmdline);

    interfaces[in_act->if_index].cf_bitmap |= AP_IFFLAG_BUSY;
    
    return 1;
}

int TcpSSLApClient::QueuePacket(stream_frame *in_frame) {
    int hostlen = ntohl(in_frame->frame_len);

    // Update the sequence ID
    ap_common_payload *cpay = (ap_common_payload *) in_frame->payload;
    cpay->station_id = (uint16_t) htons(station_id);
    cpay->sequence_id = kis_hton64((uint64_t) tx_seqid++);
   
    // Queue it into the write buffer
    if (write_buf->InsertDummy(FRAMELEN_HEADER + hostlen) == 0) {
        snprintf(errstr, 1024, "TCP AP client no room in write ring buffer to insert "
                 "queued packet.");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }
    write_buf->InsertData((uint8_t *) in_frame, FRAMELEN_HEADER);
    write_buf->InsertData((uint8_t *) in_frame->payload, hostlen);

    return FRAMELEN_HEADER + hostlen;
}

int TcpSSLApClient::ActivateInterfaces() {
    for (unsigned int x = 0; x < 4; x++) {
        if ((interfaces[x].dirty || interfaces[x].macdirty || 
             (interfaces[x].cf_bitmap & AP_IFFLAG_ERROR) != 0) &&
            (interfaces[x].cf_bitmap & AP_IFFLAG_CONFIGURED) != 0) {

            // Look for stuff that hasn't changed
            if (interfaces[x].channel == interfaces[x].old_channel &&
                interfaces[x].cf_bitmap == interfaces[x].old_cf_bitmap &&
                (interfaces[x].cf_bitmap & AP_IFFLAG_ERROR) == 0) {
                interfaces[x].dirty = 0;
                continue;
            }

            printf("debug - working on interface %d, bm %d obm %d dirty %d mdirty %d ch %d och %d\n",
                   x, interfaces[x].cf_bitmap, interfaces[x].old_cf_bitmap, interfaces[x].dirty,
                   interfaces[x].macdirty, interfaces[x].channel, interfaces[x].old_channel);

            // Change the MAC if we need to
            if (interfaces[x].macdirty) {
                // This isn't supported
                interfaces[x].macdirty = 0;
#if 0
                if (interfaces[x].old_cf_bitmap & AP_IFFLAG_ACCESSPOINT) {
                    printf("debug - Interface %u is changing BSSID, send "
                           "broadcast discconect to get users off it\n", x);
                    if (Iwconfig_Set_IntPriv(interfaces[x].ifname, "maccmd", 
                                             4, 0, errstr) < 0) {
                        // Inject it into the msgbus and then try to keep going, we'll
                        // let the actual mode change flag an error state
                        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                    }
                }
                if (Ifconfig_Set_Hwaddr(interfaces[x].ifname, errstr,
                                        interfaces[x].hwaddr) < 0) {
                    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                    interfaces[x].cf_bitmap |= AP_IFFLAG_ERROR;
                } else {
                    interfaces[x].macdirty = 0;
                }
#endif
                
            }

            // Look for stuff thats only changed channel
            // Always do a mode change
#if 0
            if (interfaces[x].channel != interfaces[x].old_channel &&
                interfaces[x].cf_bitmap == interfaces[x].old_cf_bitmap) {
                interfaces[x].dirty = 0;
                printf("debug - change channel from %u to %u but keep mode on %u\n", 
                       interfaces[x].old_channel, interfaces[x].channel, x);
                if (Iwconfig_Set_Channel(interfaces[x].ifname,
                                         interfaces[x].channel,
                                         errstr) < 0) {
                    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                    // Flag it as broken
                    interfaces[x].cf_bitmap |= AP_IFFLAG_ERROR;
                }
                continue;
            }
#endif

            // Otherwise change mode
            // Handle cleaning old modes
            
            if ((interfaces[x].old_cf_bitmap & AP_IFFLAG_SNIFFER) && 
                (interfaces[x].cf_bitmap & AP_IFFLAG_SNIFFER) == 0) {
                if (interfaces[x].kismet_pid == 0)
                    continue;

                printf("debug - Interface %u will no longer be a sniffer, kill "
                       "Kismet off\n", x);
                if (interfaces[x].kisclient != NULL) {
                    interfaces[x].kisclient->KillConnection();
                }
                kill(interfaces[x].kismet_pid, SIGKILL);
                interfaces[x].kismet_pid = 0;
            }

            if ((interfaces[x].old_cf_bitmap & AP_IFFLAG_ACCESSPOINT) &&
                (interfaces[x].cf_bitmap & AP_IFFLAG_ACCESSPOINT) == 0) {
                printf("debug - Interface %u will no longer be an AP, send "
                       "broadcast discconect to get users off it\n", x);
                if (Iwconfig_Set_IntPriv(interfaces[x].ifname, "maccmd", 
                                         4, 0, errstr) < 0) {
                    // Inject it into the msgbus and then try to keep going, we'll
                    // let the actual mode change flag an error state
                    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                }

            }

            // Set the new modes
            if (interfaces[x].cf_bitmap & AP_IFFLAG_ACCESSPOINT) {
                printf("debug - Turning interface %u into an AP on channel %u\n",
                       x, interfaces[x].channel);
                // Set the SSID
                if (Iwconfig_Set_SSID(interfaces[x].ifname, errstr, ssid) < 0) {
                    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                    interfaces[x].cf_bitmap |= AP_IFFLAG_ERROR;
                    continue;
                }
                // Set the channel
                if (Iwconfig_Set_Channel(interfaces[x].ifname, 
                                         interfaces[x].channel, errstr) < 0) {
                    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                    interfaces[x].cf_bitmap |= AP_IFFLAG_ERROR;
                    continue;
                }

                // Execute helper command to set this up
                char hlpcmdline[1024];
                snprintf(hlpcmdline, 1024, "%s %s up", aphelperpath,
                         interfaces[x].ifname);
                // ExecSysCmd(hlpcmdline);
                if (RunSysCmd(hlpcmdline, errstr) < 0)
                    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);

            } else if (interfaces[x].cf_bitmap & AP_IFFLAG_SNIFFER) {
                if (interfaces[x].kismet_pending || interfaces[x].kismet_pid != 0)
                    continue;

                printf("debug - Turning interface %u into a sniffer, have to spawn "
                       "Kismet and attach a sniffer to it.\n", x);
                
                // Remove us from the bridge - raw 802.11 kills it
                char hlpcmdline[1024];
                snprintf(hlpcmdline, 1024, "%s %s down", aphelperpath,
                         interfaces[x].ifname);
                if (RunSysCmd(hlpcmdline, errstr) < 0)
                    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);

                // Form a kismet command line and start it
                char kiscmdline[1024];
                snprintf(kiscmdline, 1024, "%s -c hostap,%s,spawn%u -p %u -n -s",
                         kispath, interfaces[x].ifname, kisport, kisport);
                interfaces[x].kismet_pid = ExecSysCmd(kiscmdline);
                printf("debug - just execed kismet, I am %d\n", getpid());
                interfaces[x].kismet_port = kisport;
                kisport++;

                snprintf(errstr, 1024, "Spawned Kismet pid %u with cmdline: %s",
                         interfaces[x].kismet_pid, kiscmdline);
                msgbus->InjectMessage(errstr, MSGFLAG_INFO);

                interfaces[x].kismet_failedcon = 0;
                
                TcpSSLApClientKismetTimerRec *tr = new TcpSSLApClientKismetTimerRec;
                tr->cli = this;
                tr->ifnum = x;

                timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * 4,
                                           NULL, 0, 
                                           &SSLApClientConnectKismetTimer, 
                                           (void *) tr);
                interfaces[x].kismet_pending = 1;

            } else if (interfaces[x].cf_bitmap & AP_IFFLAG_RAWTX) {
                // Remove us from the bridge so we can't clobber anything
                char hlpcmdline[1024];
                snprintf(hlpcmdline, 1024, "%s %s down", aphelperpath,
                         interfaces[x].ifname);
                if (RunSysCmd(hlpcmdline, errstr) < 0)
                    msgbus->InjectMessage(errstr, MSGFLAG_ERROR);

                
                printf("debug - Turning interface %u into a rawtx agent\n", x);
            }

            interfaces[x].dirty = 0;
            interfaces[x].old_cf_bitmap = interfaces[x].cf_bitmap;

        }

    }

    return 1;
}

int TcpSSLApClient::ConnectKismetInterface(unsigned int in_ifnum) {
    // Disconnect and reconnect if one exists, or make one
    if (interfaces[in_ifnum].kisclient != NULL) {
        interfaces[in_ifnum].kisclient->KillConnection();
    } else {
        interfaces[in_ifnum].kisclient = new TcpKismetClient;
        interfaces[in_ifnum].kisclient->RegisterMessageBus(msgbus);
        interfaces[in_ifnum].kisclient->RegisterTimetracker(timetracker);
    }

    // Connect to the local machine
    if (interfaces[in_ifnum].kisclient->Connect("localhost", 
                                                interfaces[in_ifnum].kismet_port) < 0) {
        interfaces[in_ifnum].kismet_failedcon++;

        // If we can't connect after 5 attempts, assume something went wonky with the 
        // kismet process, set us to not pending and dirty so we get respawned, then
        // activate the interfaces again
        if (interfaces[in_ifnum].kismet_failedcon > 5) {
            snprintf(errstr, 1024, "Failed to connect to Kismet 5 times, forcing "
                     "respawn of Kismet process.");
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            interfaces[in_ifnum].kismet_pending = 0;
            interfaces[in_ifnum].dirty = 1;

            if (interfaces[in_ifnum].kismet_pid != 0)
                kill(interfaces[in_ifnum].kismet_pid, SIGKILL);
            interfaces[in_ifnum].kismet_pid = 0;

            ActivateInterfaces();
            return 1;
        }
        
        return -1;
    }
    
    interfaces[in_ifnum].kismet_time = time(0);

    // Unset any errors
    interfaces[in_ifnum].cf_bitmap &= ~AP_IFFLAG_ERROR;

    // Unset pending
    interfaces[in_ifnum].kismet_pending = 0;
    
    return 1;
}

int TcpSSLApClient::ThrashTick() {
    int nfail = 0;

    for (unsigned int x = 0; x < 4; x++) {
        if ((interfaces[x].cf_bitmap & AP_IFFLAG_ERROR) == 0) {
            interfaces[x].failures = 0;
        } else {
            interfaces[x].failures++;
            
            // Start generating errors
            if (interfaces[x].failures >= 10) {
                snprintf(errstr, 1024, "Interface '%s' appears to be in trouble, "
                         "in error state for %d seconds.", interfaces[x].ifname,
                         interfaces[x].failures);
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            } 

            nfail++;
        }
    }

    // Try to reconfigure borked ones
    if (nfail > 0) 
        ActivateInterfaces();

    return 1;
}

