/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "tcpsslclient.h"

TcpSSLClient::TcpSSLClient() : TcpClient() {
    SSL_library_init();
    SSL_load_error_strings();

    RAND_load_file("/dev/urandom", 1024);

    // Build the SSL contexts and link them
    con_ctx = SSL_CTX_new(SSLv23_method());
    SSL_CTX_set_options(con_ctx, SSL_OP_ALL | SSL_OP_NO_SSLv2);
}

TcpSSLClient::~TcpSSLClient() {
    KillConnection();
    SSL_CTX_free(con_ctx);
}

char tcpcli_ctx_pem_passphrase[1024];

int tcpcli_pem_passwd_cb(char *buf, int size, int rwflag, void *password) {
    strncpy(buf, (char *)(tcpcli_ctx_pem_passphrase), size);
    buf[size - 1] = '\0';
    return (strlen(buf));
}

int TcpSSLClient::SetCTXCert(const char *in_certfile, const char *in_passphrase) {
    snprintf(tcpcli_ctx_pem_passphrase, 1024, "%s", in_passphrase);

    SSL_CTX_set_default_passwd_cb(con_ctx, &tcpcli_pem_passwd_cb);

    if (SSL_CTX_use_certificate_chain_file(con_ctx, in_certfile) != 1) {
        snprintf(errstr, 1024, "TCP SSL client unable to load certificate chain "
                 "file %s", in_certfile);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (SSL_CTX_use_PrivateKey_file(con_ctx, in_certfile, SSL_FILETYPE_PEM) != 1) {
        snprintf(errstr, 1024, "TCP SSL client unable to load PEM private key %s",
                 in_certfile);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    memset(tcpcli_ctx_pem_passphrase, 0, 1024);

    return 1;
}

int TcpSSLClient::SetCTXVerify(const char *in_cafile, const char *in_cadir) {
    if (SSL_CTX_load_verify_locations(con_ctx, in_cafile, in_cadir) != 1) {
        snprintf(errstr, 1024, "TCP SSL client unable to load certificate "
                 "authority data from %s%s%s",
                 in_cafile == NULL ? "" : in_cafile,
                 in_cafile == NULL ? "" : " ",
                 in_cadir == NULL ? "" : in_cadir);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }
    SSL_CTX_set_verify(con_ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
                       &ssl_common_cert_verify_cb);
    SSL_CTX_set_verify_depth(con_ctx, 10);

    return 1;
}

int TcpSSLClient::SetCTXCipherList(const char *in_cipherlist) {
    if (SSL_CTX_set_cipher_list(con_ctx, in_cipherlist) != 1) {
        snprintf(errstr, 1024, "TCP SSL client unable to set cipher list \"%s\"",
                 in_cipherlist);
        return -1;
    }
    
    return 1;
}

unsigned int TcpSSLClient::MergeSet(fd_set in_rset, fd_set in_wset, 
                                    unsigned int in_max_fd,
                                    fd_set *out_rset, fd_set *out_wset) {
    unsigned int max;

    FD_ZERO(out_rset);
    FD_ZERO(out_wset);

    if ((int) in_max_fd < client_fd && cl_valid)
        max = client_fd;
    else
        max = in_max_fd;

    for (unsigned int x = 0; x <= max; x++) {
        if (FD_ISSET(x, &in_rset))
            FD_SET(x, out_rset);
        if (FD_ISSET(x, &in_wset))
            FD_SET(x, out_wset);
    }

    if (cl_valid) {
        // Don't try to read if we're blocked into a write
        if (ssl_blocked_write == 0)
            FD_SET(client_fd, out_rset);

        // Don't look to write if we're blocked into a read
        if (write_buf->FetchLen() > 0 && ssl_blocked_read == 0)
            FD_SET(client_fd, out_wset);
    }

    return max;
}

int TcpSSLClient::KillConnection() {
    if (cl_valid) {
        SSL_shutdown(con_ssl);
        SSL_free(con_ssl);

        TcpClient::KillConnection();
    }

    return 1;
}

int TcpSSLClient::Connect(const char *in_remotehost, short int in_port) {
    int ret;
    int ssl_err;

    // Call the standard TCP stuff, with blocking enabled
    connect_nonblocking = 0;
    if ((ret = TcpClient::Connect(in_remotehost, in_port)) < 0)
        return ret;

    con_ssl = SSL_new(con_ctx);
    SSL_set_mode(con_ssl, (SSL_MODE_ENABLE_PARTIAL_WRITE | 
                           SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER));

    SSL_set_fd(con_ssl, client_fd);

    int save_mode = fcntl(client_fd, F_GETFL, 0);
    fcntl(client_fd, F_SETFL, save_mode & ~O_NONBLOCK);

    if ((ret = SSL_connect(con_ssl)) <= 0) {
        ssl_err = SSL_get_error(con_ssl, ret);
    }

    fcntl(client_fd, F_SETFL, save_mode);

    ssl_busy_read = 0;
    ssl_busy_write = 0;
    ssl_blocked_read = 0;
    ssl_blocked_write = 0;

    X509 *peercert;
    if (!(peercert = SSL_get_peer_certificate(con_ssl))) {
        snprintf(errstr, 1024, "TCP SSL client unable to get peer certificate");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        // Hack in a local debug output
        ERR_print_errors_fp(stderr);
        return -1;
    }

    if (X509_cmp_current_time(X509_get_notBefore(peercert)) >= 0) {
        snprintf(errstr, 1024, "TCP SSL server got peer cert from the future");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (X509_cmp_current_time(X509_get_notAfter(peercert)) <= 0) {
        snprintf(errstr, 1024, "TCP SSL server got expired peer cert");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    printf("debug - more server auth to be done\n");
    X509_free(peercert);

    // Set nonblocking mode
    if (SetNonblocking() < 0)
        return -1;
    
    return 1;
}

int TcpSSLClient::ReadBytes() {
    uint8_t recv_bytes[1024];
    int ret;

    if (cl_valid == 0)
        return -1;
   
    // If we woke up able to service a read and we're busy in a write, go
    // handle it instead
    if (ssl_busy_write) {
        return WriteBytes();
    }

    ssl_busy_read = 0;
    ssl_busy_write = 0;
    ssl_blocked_read = 0;
    ssl_blocked_write = 0;

    if ((ret = SSL_read(con_ssl, recv_bytes, 1024)) <= 0) {
        switch (SSL_get_error(con_ssl, ret)) {
            case SSL_ERROR_ZERO_RETURN:
                KillConnection();
                snprintf(errstr, 1024, "SSL client connection shut down");
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                return 0;
                break;
            case SSL_ERROR_WANT_READ:
                // We're locked into a read block
                ssl_busy_read = 1;
                ssl_blocked_read = 1;
                printf("debug - client blocked into a read\n");
                return 0;
                break;
            case SSL_ERROR_WANT_WRITE:
                // Blocked into write
                ssl_busy_read = 1;
                ssl_blocked_write = 1;
                printf("debug - client blocked into a write\n");
                return 0;
                break;
            default:
                snprintf(errstr, 1024, "SSL client unhandled SSL read error %d",
                         SSL_get_error(con_ssl, ret));
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                return -1;
                break;
        }
    } else {
        if (read_buf->InsertData(recv_bytes, ret) == 0) {
            snprintf(errstr, 1024, "TCP client read error, ring buffer full");
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            return -1;
        }

    }

    return ret;
}

int TcpSSLClient::WriteBytes() {
    uint8_t dptr[1024];
    int dlen, ret;
   
    // If we're waiting for a read and we woke up being able to write, go service it
    if (ssl_busy_read) {
        return ReadBytes();
    }

    ssl_blocked_read = 0;
    ssl_blocked_write = 0;
    ssl_busy_read = 0;
    ssl_busy_write = 0;

    write_buf->FetchPtr(dptr, 1024, &dlen);

    if ((ret = SSL_write(con_ssl, dptr, dlen)) <= 0) {
        switch (SSL_get_error(con_ssl, ret)) {
            case SSL_ERROR_ZERO_RETURN:
                KillConnection();
                snprintf(errstr, 1024, "SSL client connection shut down");
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                return 0;
                break;
            case SSL_ERROR_WANT_READ:
                ssl_busy_write = 1;
                ssl_blocked_read = 1;
                printf("debug - client blocked into read during write\n");
                return 0;
                break;
            case SSL_ERROR_WANT_WRITE:
                ssl_busy_write = 1;
                ssl_blocked_write = 1;
                printf("debug - client blocked into a write\n");
                return 0;
                break;
            default:
                snprintf(errstr, 1024, "SSL client unhandled SSL write error %d",
                         SSL_get_error(con_ssl, ret));
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                return -1;
                break;
        }
    } else {
        write_buf->MarkRead(ret);
    }

    return ret;
}

