/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "tcpsslserver.h"

TcpSSLServer::TcpSSLServer() : TcpServer() {
    SSL_library_init();
    SSL_load_error_strings();

    RAND_load_file("/dev/urandom", 1024);

    // Make the context
    serv_ctx = SSL_CTX_new(SSLv23_method());
    SSL_CTX_set_options(serv_ctx, SSL_OP_ALL | SSL_OP_NO_SSLv2);
}

TcpSSLServer::~TcpSSLServer() {
    Shutdown();
    SSL_CTX_free(serv_ctx);
}

char tcpsrv_ctx_pem_passphrase[1024];

int tcpssl_pem_passwd_cb(char *buf, int size, int rwflag, void *password) {
    strncpy(buf, (char *)(tcpsrv_ctx_pem_passphrase), size);
    buf[size - 1] = '\0';
    return(strlen(buf));
}

int TcpSSLServer::SetCTXCert(const char *in_certfile, const char *in_passphrase) {
    // Copy pw into our own buffer
    snprintf(tcpsrv_ctx_pem_passphrase, 1024, "%s", in_passphrase);
   
    // Set the password callback for our context
    SSL_CTX_set_default_passwd_cb(serv_ctx, &tcpssl_pem_passwd_cb);

    // Load the PEMs
    if (SSL_CTX_use_certificate_chain_file(serv_ctx, in_certfile) != 1) {
        snprintf(errstr, 1024, "TCP SSL server unable to load certificate chain "
                 "file %s", in_certfile);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (SSL_CTX_use_PrivateKey_file(serv_ctx, in_certfile, SSL_FILETYPE_PEM) != 1) {
        snprintf(errstr, 1024, "TCP SSL server unable to load PEM private key %s",
                 in_certfile);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    // Zero the stored passphrase
    memset(tcpsrv_ctx_pem_passphrase, 0, 1024);
    
    return 1;
}

int TcpSSLServer::SetCTXVerify(const char *in_cafile, const char *in_cadir) {
    if (SSL_CTX_load_verify_locations(serv_ctx, in_cafile, in_cadir) != 1) {
        snprintf(errstr, 1024, "TCP SSL server unable to load certificate "
                 "authority data from %s%s%s",
                 in_cafile == NULL ? "" : in_cafile,
                 in_cafile == NULL ? "" : " ",
                 in_cadir == NULL ? "" : in_cadir);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }
    SSL_CTX_set_verify(serv_ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
                       &ssl_common_cert_verify_cb);
    SSL_CTX_set_verify_depth(serv_ctx, 10);

    return 1;
}

int TcpSSLServer::SetCTXCipherList(const char *in_cipherlist) {
    if (SSL_CTX_set_cipher_list(serv_ctx, in_cipherlist) != 1) {
        snprintf(errstr, 1024, "TCP SSL server unable to set cipher list \"%s\"",
                 in_cipherlist);
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    return 1;
}   

void TcpSSLServer::KillConnection(int in_fd) {
    // printf("debug - ssl kill connection %d\n", in_fd);

    map<int, tcpssl_conrec *>::iterator sitr = ssl_rec_map.find(in_fd);
    if (sitr != ssl_rec_map.end()) {
        SSL_shutdown(sitr->second->ssl_con);
        SSL_free(sitr->second->ssl_con);

        if (sitr->second->ssl_peercert != NULL)
            X509_free(sitr->second->ssl_peercert);

        delete(sitr->second);
        ssl_rec_map.erase(sitr);
    }
    
    TcpServer::KillConnection(in_fd);
}

unsigned int TcpSSLServer::MergeSet(fd_set in_rset, fd_set in_wset,
                                 unsigned int in_max_fd,
                                 fd_set *out_rset, fd_set *out_wset) {
    unsigned int max;

    FD_ZERO(out_rset);
    FD_ZERO(out_wset);
   
    if (in_max_fd < max_fd) {
        max = max_fd;
    } else {
        max = in_max_fd;
    }

    // Set the server fd
    FD_SET(serv_fd, out_rset);
    
    for (unsigned int x = 0; x <= max; x++) {
        // If we're not blocking on a write, it's ok to try to read from
        // this fd...
        if (FD_ISSET(x, &in_rset) || (FD_ISSET(x, &server_fdset) &&
                                      ssl_rec_map[x]->ssl_blocked_write == 0))
            FD_SET(x, out_rset);

        // We can't write data if they're blocked into a read
        if (FD_ISSET(x, &in_wset) || 
            (write_buf_map.find(x) != write_buf_map.end() &&
             write_buf_map[x]->FetchLen() > 0 && 
             ssl_rec_map[x]->ssl_blocked_read == 0))
            FD_SET(x, out_wset);
    }
   
    return max;
}

int TcpSSLServer::Validate(int in_fd) {
    // This whole thing is a little "funny".  It gets called by the TCP
    // accept, sets up the SSL connections, and then sets up the SSL nonblocking 
    // accept to try to establish the full connection.. This is probably bad design.
    
    // IP validate
    if (ValidateIPFilter(in_fd) < 0)
        return -1;

    // Make a new SSL record tied to this preexisting socket
    tcpssl_conrec *crec = new tcpssl_conrec;
    int ret;

    // Create a SSL connection in our server context and bind it to the
    // socket we justaccepted
    crec->ssl_con = SSL_new(serv_ctx);
    // printf("debug - ssl new off server context worked\n");
    SSL_set_fd(crec->ssl_con, in_fd);
    // printf("debug - ssl fd set to %d\n", in_fd);

    crec->ssl_established = 0;
    crec->ssl_busy_read = 0;
    crec->ssl_busy_write = 0;
    crec->ssl_blocked_read = 0;
    crec->ssl_blocked_write = 0;
    crec->ssl_peercert = NULL;

    ssl_rec_map[in_fd] = crec;
    
    // SSL-level accept
    // printf("debug - ssl nonblock accept call for %d\n", in_fd);
    if ((ret = SSLNonBlockAccept(in_fd)) < 0) {
        // printf("debug - ssl nba lt 0: %d\n", ret);
        return -1;
    }

    return 1;
}

int TcpSSLServer::SSLNonBlockAccept(int in_fd) {
    tcpssl_conrec *crec = ssl_rec_map[in_fd];
    int ret = 0;
    int ssl_err = 0;

    if ((ret = SSL_accept(crec->ssl_con)) <= 0) {
        // printf("debug - errno %d %s\n", errno, strerror(errno));
        ssl_err = SSL_get_error(crec->ssl_con, ret);

        if (SSL_get_shutdown(crec->ssl_con) & SSL_RECEIVED_SHUTDOWN) {
            // printf("debug - ssl accept received shutdown\n");
            snprintf(errstr, 1024, "TCP SSL server ssl accept and validate failed");
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            KillConnection(in_fd);
            // Hack in a local debug output
            ERR_print_errors_fp(stderr);
            return -1;
        } else if (ssl_err == SSL_ERROR_WANT_READ || ssl_err == SSL_ERROR_WANT_WRITE) {
            // printf("debug - ssl want read/write\n");
            return 0;
        } else {
            // printf("debug - ssl ret %d error %d %s\n", ret, ssl_err, strerror(errno));
            snprintf(errstr, 1024, "TCP SSL server ssl accept and validate failed");
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            KillConnection(in_fd);
            return -1;
        }
    } 
        // printf("debug - SSL_accept returned %d\n", ret);

    // SSL accep worked
    crec->ssl_established = 1;

    // X509 *peercert;
    if (!(crec->ssl_peercert = SSL_get_peer_certificate(crec->ssl_con))) {
        snprintf(errstr, 1024, "TCP SSL server unable to get peer certificate");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        crec->ssl_peercert = NULL;
        return -1;
    }

    if (X509_cmp_current_time(X509_get_notBefore(crec->ssl_peercert)) >= 0) {
        snprintf(errstr, 1024, "TCP SSL server got peer cert from the future");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    if (X509_cmp_current_time(X509_get_notAfter(crec->ssl_peercert)) <= 0) {
        snprintf(errstr, 1024, "TCP SSL server got expired peer cert");
        msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
        return -1;
    }

    printf("debug - more peer auth to be done here.\n");
    // X509_free(peercert);

    // Set the SSL nonblocking modes
    SSL_set_mode(crec->ssl_con, (SSL_MODE_ENABLE_PARTIAL_WRITE | 
                                 SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER));

    return 1;
}

int TcpSSLServer::ReadBytes(int in_fd) {
    uint8_t recv_bytes[1024];
    int ret;
    tcpssl_conrec *crec = ssl_rec_map[in_fd];

    // If we're trying to write and the read flag is set, then we can try to
    // complete our write - go do that.
    if (crec->ssl_busy_write) {
        // printf("debug - waiting for busy write, going to write\n");
        return WriteBytes(in_fd);
    }
    
    // Don't do error checking here because we should never get called
    // if we're not ready to do work
    
    // Try to accept this connection if its stuck in an async blocked accept
    if (crec->ssl_established == 0) {
        // printf("debug - not establshed, nonblock accept try before read\n");
        if ((ret = SSLNonBlockAccept(in_fd)) <= 0)
            return ret;
    }
 
    // Zero the state
    crec->ssl_blocked_read = 0;
    crec->ssl_blocked_write = 0;
    crec->ssl_busy_read = 0;
    crec->ssl_busy_write = 0;
    
    if ((ret = SSL_read(crec->ssl_con, recv_bytes, 1024)) <= 0) {
        switch (SSL_get_error(crec->ssl_con, ret)) {
            case SSL_ERROR_ZERO_RETURN:
                // Connection dead, kill it and return a 0
                KillConnection(in_fd);
                snprintf(errstr, 1024, "SSL server client fd %d shut down", in_fd);
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                return 0;
                break;
            case SSL_ERROR_WANT_READ:
                // We're locked into a read block, this means we can't do anything
                // but read next time
                crec->ssl_busy_read = 1;
                crec->ssl_blocked_read = 1;
                // printf("debug - client %d blocked into a read\n", in_fd);
                return 0;
                break;
            case SSL_ERROR_WANT_WRITE:
                // We can't read until we've written something, block into a
                // write
                crec->ssl_busy_read = 1;
                crec->ssl_blocked_write = 1;
                // printf("debug - client %d blocked into a write\n", in_fd);
                return 0;
                break;
            default:
                snprintf(errstr, 1024, "SSL server client fd %d unhandled SSL "
                         "error %d", in_fd, SSL_get_error(crec->ssl_con, ret));
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                KillConnection(in_fd);
                return -1;
                break;
        }
    } else {
        if (read_buf_map[in_fd]->InsertData(recv_bytes, ret) == 0) {
            snprintf(errstr, 1024, "TCP server client fd %d read error, ring "
                     "buffer full", in_fd);
            msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
            return -1;
        }
    }

    return ret;
}

int TcpSSLServer::WriteBytes(int in_fd) {
    uint8_t dptr[1024];
    int dlen, ret;
    tcpssl_conrec *crec = ssl_rec_map[in_fd];

    // If we're trying to read and the write flag is set, then we can try to
    // complete our read - go do that.
    if (crec->ssl_busy_read) {
        return ReadBytes(in_fd);
    }
    
    // Try to accept this connection if its stuck in an async blocked accept
    if (crec->ssl_established == 0) {
        // printf("debug - not establshed, nonblock accept try before read\n");
        if ((ret = SSLNonBlockAccept(in_fd)) <= 0)
            return ret;
    }
 
    // Zero the state
    crec->ssl_blocked_read = 0;
    crec->ssl_blocked_write = 0;
    crec->ssl_busy_read = 0;
    crec->ssl_busy_write = 0;

    write_buf_map[in_fd]->FetchPtr(dptr, 1024, &dlen);

    if ((ret = SSL_write(crec->ssl_con, dptr, dlen)) <= 0) {
        switch (SSL_get_error(crec->ssl_con, ret)) {
            case SSL_ERROR_ZERO_RETURN:
                // Connection dead, kill it and return a 0
                KillConnection(in_fd);
                snprintf(errstr, 1024, "SSL server client fd %d shut down", in_fd);
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                return 0;
                break;
            case SSL_ERROR_WANT_READ:
                // We're locked into a read block, this means we can't do anything
                // but read next time
                crec->ssl_busy_write = 1;
                crec->ssl_blocked_read = 1;
                // printf("debug - write client %d blocked into a read\n", in_fd);
                return 0;
                break;
            case SSL_ERROR_WANT_WRITE:
                // We can't read until we've written something, block into a
                // write
                crec->ssl_busy_write = 1;
                crec->ssl_blocked_write = 1;
                // printf("debug - write client %d blocked into a write\n", in_fd);
                return 0;
                break;
            default:
                snprintf(errstr, 1024, "SSL server client fd %d unhandled SSL "
                         "error %d", in_fd, SSL_get_error(crec->ssl_con, ret));
                msgbus->InjectMessage(errstr, MSGFLAG_ERROR);
                KillConnection(in_fd);
                return -1;
                break;
        }
    } else {
        write_buf_map[in_fd]->MarkRead(ret);
    }

    return ret;
}


